import functools
import inspect


# credits: https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3/25959545#25959545
def get_class_that_defined_method(meth):
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth):
        for c in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in c.__dict__:
                return c
        meth = getattr(meth, "__func__", meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        c = getattr(
            inspect.getmodule(meth),
            meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0],
            None,
        )
        if isinstance(c, type):
            return c
    return getattr(meth, "__objclass__", None)  # handle special descriptor objects



from __future__ import annotations

from functools import lru_cache, reduce
from operator import or_
from typing import TYPE_CHECKING, Iterator, List, Set, Tuple, Union

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcBlock
from woke.ir.enums import ModifiesStateFlag
from woke.ir.statements.abc import StatementAbc
from woke.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from ..declarations.function_definition import FunctionDefinition
    from ..declarations.modifier_definition import ModifierDefinition
    from ..meta.try_catch_clause import TryCatchClause
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class Block(StatementAbc):
    """
    Block statements group multiple statements into a single block.

    !!! example
        Lines 3-5 in the following code:
        ```solidity linenums="1"
        contract Foo {
            function bar(uint a, uint b) public pure returns(uint)
            {
                return a + b;
            }
        }
        ```
    """

    _ast_node: SolcBlock
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,  # statements
        FunctionDefinition,
        ModifierDefinition,  # declarations
        TryCatchClause,  # meta
    ]

    _statements: List[StatementAbc]

    def __init__(self, init: IrInitTuple, block: SolcBlock, parent: SolidityAbc):
        super().__init__(init, block, parent)
        self._statements = [
            StatementAbc.from_ast(init, statement, self)
            for statement in block.statements
        ]

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        if self._statements is not None:
            for statement in self._statements:
                yield from statement

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
        FunctionDefinition,
        ModifierDefinition,
        TryCatchClause,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def statements(self) -> Tuple[StatementAbc, ...]:
        """
        Can be empty.
        Returns:
            Statements in the block.
        """
        return tuple(self._statements)

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        if self.statements is None:
            return set()
        return reduce(
            or_,
            (statement.modifies_state for statement in self.statements),
            set(),
        )

    def statements_iter(self) -> Iterator["StatementAbc"]:
        yield self
        if self._statements is not None:
            for statement in self._statements:
                yield from statement.statements_iter()



import json
import platform
import time

from .abc import ProtocolAbc

if platform.system() == "Windows":
    import win32file  # pyright: ignore reportMissingModuleSource

    class IpcProtocol(ProtocolAbc):  # pyright: ignore reportGeneralTypeIssues
        _uri: str
        _timeout: float

        def __init__(self, uri: str, timeout: float):
            self._uri = uri
            self._timeout = timeout

        def __enter__(self):
            self._handle = win32file.CreateFile(
                self._uri,
                win32file.GENERIC_READ | win32file.GENERIC_WRITE,
                0,
                None,
                win32file.OPEN_EXISTING,
                0,
                None,
            )

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._handle.close()

        def send_recv(self, data: str):
            win32file.WriteFile(
                self._handle,  # pyright: ignore reportGeneralTypeIssues
                data.encode("utf-8"),
            )
            received = bytearray()
            start = time.perf_counter()

            while time.perf_counter() - start < self._timeout:
                res, data = win32file.ReadFile(
                    self._handle, 4096  # pyright: ignore reportGeneralTypeIssues
                )
                received += data  # pyright: ignore reportGeneralTypeIssues
                if not received.rstrip().endswith((b"}", b"]")):
                    continue
                try:
                    return json.loads(received.decode("utf-8"))
                except json.JSONDecodeError:
                    continue
            raise TimeoutError("IPC communication timeout")

else:
    import socket

    class IpcProtocol(ProtocolAbc):
        _uri: str
        _timeout: float
        _socket: socket.socket

        def __init__(self, uri: str, timeout: float):
            self._uri = uri
            self._timeout = timeout

        def __enter__(self):
            self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._socket.connect(self._uri)
            self._socket.settimeout(0.0005)

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._socket.close()

        def send_recv(self, data: str):
            self._socket.sendall(data.encode("utf-8"))
            received = bytearray()
            start = time.perf_counter()

            while time.perf_counter() - start < self._timeout:
                try:
                    received += self._socket.recv(4096)
                except socket.timeout:
                    if not received.rstrip().endswith((b"}", b"]")):
                        continue
                    try:
                        return json.loads(received.decode("utf-8"))
                    except json.JSONDecodeError:
                        continue
            raise TimeoutError("IPC communication timeout")



from typing import Iterator, Set, Tuple

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcNewExpression
from woke.ir.enums import ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc
from woke.ir.type_names.abc import TypeNameAbc
from woke.ir.types import Contract
from woke.ir.utils import IrInitTuple


class NewExpression(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcNewExpression
    _parent: SolidityAbc  # TODO: make this more specific

    _type_name: TypeNameAbc

    def __init__(
        self, init: IrInitTuple, new_expression: SolcNewExpression, parent: SolidityAbc
    ):
        super().__init__(init, new_expression, parent)
        self._type_name = TypeNameAbc.from_ast(init, new_expression.type_name, self)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._type_name

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def type_name(self) -> TypeNameAbc:
        return self._type_name

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        if isinstance(self.type, Contract):
            return {(self, ModifiesStateFlag.DEPLOYS_CONTRACT)}
        else:
            return set()



from __future__ import annotations

import logging
from collections import deque
from functools import reduce
from operator import or_
from typing import List, Optional, Set, Tuple, Union

import networkx as nx

import woke.ir as ir
import woke.ir.types as types
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)

logger = logging.getLogger(__name__)


def address_is_safe(addr: ir.VariableDeclaration) -> Optional[bool]:
    from woke.analysis.ownable import variable_is_owner

    if variable_is_owner(addr):
        return True

    if isinstance(addr.parent, ir.ParameterList) and (
        isinstance(addr.parent.parent, ir.FunctionDefinition)
    ):
        function_definition = addr.parent.parent
        assert isinstance(function_definition, ir.FunctionDefinition)
        if function_definition.visibility in {
            ir.enums.Visibility.PUBLIC,
            ir.enums.Visibility.EXTERNAL,
        }:
            return False

        return None
    elif addr.is_state_variable:
        if addr.mutability in {
            ir.enums.Mutability.CONSTANT,
            ir.enums.Mutability.IMMUTABLE,
        }:
            return True
        return False
    elif isinstance(addr.parent, ir.VariableDeclarationStatement):
        return None
    else:
        logger.warning(
            f"Unable to detect if address source is safe: {addr.parent}\n{addr.source}"
        )
        return None


def find_low_level_call_source_address(
    expression: ir.ExpressionAbc,
) -> Optional[
    Union[
        ir.ContractDefinition,
        ir.VariableDeclaration,
        ir.Literal,
        ir.enums.GlobalSymbolsEnum,
    ]
]:
    t = expression.type
    if isinstance(t, types.Type):
        if not isinstance(t.actual_type, types.Contract):
            return None
    elif not isinstance(expression.type, (types.Address, types.Contract)):
        return None

    while True:
        if isinstance(expression, (ir.Identifier, ir.Literal)):
            break
        elif isinstance(expression, ir.FunctionCall):
            if expression.kind == ir.enums.FunctionCallKind.FUNCTION_CALL:
                function_called = expression.function_called
                if (
                    isinstance(function_called, ir.FunctionDefinition)
                    and function_called.body is not None
                ):
                    returns = [
                        statement
                        for statement in function_called.body.statements_iter()
                        if isinstance(statement, ir.Return)
                    ]

                    if len(returns) == 1 and returns[0].expression is not None:
                        expression = returns[0].expression
                    else:
                        return None
                else:
                    logger.debug(
                        f"Unable to find source: {expression.expression}\n{expression.source}"
                    )
                    return None
            elif expression.kind == ir.enums.FunctionCallKind.TYPE_CONVERSION:
                if len(expression.arguments) != 1:
                    logger.debug(
                        f"Unable to find source: {expression}\n{expression.source}"
                    )
                    return None
                expression = expression.arguments[0]
        elif isinstance(expression, ir.MemberAccess):
            if isinstance(
                expression.referenced_declaration, ir.enums.GlobalSymbolsEnum
            ):
                return expression.referenced_declaration
            expression = expression.expression
        elif (
            isinstance(expression, ir.TupleExpression)
            and len(expression.components) == 1
            and expression.components[0] is not None
        ):
            expression = expression.components[0]
        else:
            logger.debug(f"Unable to find source: {expression}\n{expression.source}")
            return None

    t = expression.type
    if isinstance(t, types.Type):
        if not isinstance(t.actual_type, types.Contract):
            return None
    elif not isinstance(expression.type, (types.Address, types.Contract)):
        logger.debug(f"Unable to find source: {expression.source}")
        return None

    if isinstance(expression, ir.Literal):
        return expression

    assert isinstance(expression, ir.Identifier)
    referenced_declaration = expression.referenced_declaration
    if not isinstance(
        referenced_declaration,
        (ir.ContractDefinition, ir.VariableDeclaration, ir.enums.GlobalSymbolsEnum),
    ):
        logger.debug(
            f"Unable to find source:\n{expression.source}\n{expression.parent.source}"
        )
        return None
    return referenced_declaration


def _modifies_state_after_statement(
    function_definition: ir.FunctionDefinition,
    statement: ir.StatementAbc,
) -> Set[Tuple[ir.IrAbc, ir.enums.ModifiesStateFlag]]:
    from woke.analysis.cfg import CfgBlock

    ret: Set[Tuple[ir.IrAbc, ir.enums.ModifiesStateFlag]] = set()
    cfg = function_definition.cfg
    assert cfg is not None
    start = cfg.get_cfg_block(statement)
    graph = cfg.graph

    if not nx.has_path(graph, start, cfg.end_block):
        return ret

    if statement not in start.statements:
        assert statement == start.control_statement
    else:
        index = start.statements.index(statement)
        for s in start.statements[index + 1 :]:
            ret.update(s.modifies_state)
        if start.control_statement is not None:
            ret.update(start.control_statement.modifies_state)

    queue = deque([])
    visited: Set[CfgBlock] = set()

    for _, to in graph.out_edges(start):  # pyright: ignore reportGeneralTypeIssues
        if to in visited:
            continue
        if not nx.has_path(graph, to, cfg.end_block):
            continue
        queue.append(to)

    while len(queue):
        block = queue.popleft()
        for s in block.statements:
            ret.update(s.modifies_state)
        if block.control_statement is not None:
            ret.update(block.control_statement.modifies_state)

        to: CfgBlock
        for _, to in graph.out_edges(block):  # pyright: ignore reportGeneralTypeIssues
            if to in visited:
                continue
            if not nx.has_path(graph, to, cfg.end_block):
                continue

            visited.add(to)
            queue.append(to)

    return ret


def check_reentrancy_in_function(
    function_definition: ir.FunctionDefinition,
    statement: ir.StatementAbc,
    address_source: ir.ExpressionAbc,
    child_modifies_state: Set[Tuple[ir.IrAbc, ir.enums.ModifiesStateFlag]],
    checked_statements: Set[ir.StatementAbc],
) -> List[Detection]:
    from woke.analysis.ownable import statement_is_publicly_executable
    from woke.analysis.utils import (
        get_all_base_and_child_functions,
        pair_function_call_arguments,
    )

    # TODO check non-reentrant
    if not statement_is_publicly_executable(statement, check_only_eoa=True):
        return []

    source_address_declaration = find_low_level_call_source_address(address_source)
    is_safe = None
    if source_address_declaration is None:
        logger.debug(f"{address_source.source}")
    elif isinstance(source_address_declaration, ir.enums.GlobalSymbolsEnum):
        if source_address_declaration == ir.enums.GlobalSymbolsEnum.THIS:
            is_safe = True
        elif source_address_declaration in {
            ir.enums.GlobalSymbolsEnum.MSG_SENDER,
            ir.enums.GlobalSymbolsEnum.TX_ORIGIN,
        }:
            is_safe = False
        else:
            is_safe = None
            logger.debug(f"{source_address_declaration}:")
    elif isinstance(source_address_declaration, ir.ContractDefinition):
        if source_address_declaration.kind == ir.enums.ContractKind.LIBRARY:
            is_safe = True
    elif isinstance(source_address_declaration, ir.Literal):
        is_safe = True
    else:
        is_safe = address_is_safe(source_address_declaration)

    if is_safe:
        return []

    checked_statements.add(statement)
    ret = []

    this_modifies_state = set(child_modifies_state)
    this_modifies_state.update(
        _modifies_state_after_statement(function_definition, statement)
    )

    if len(this_modifies_state) and function_definition.visibility in {
        ir.enums.Visibility.PUBLIC,
        ir.enums.Visibility.EXTERNAL,
    }:
        state_mods = reduce(or_, (mod[1] for mod in this_modifies_state))
        ret.append(
            Detection(
                statement,
                f"Exploitable from `{function_definition.canonical_name}`, address is safe: {is_safe}, state modified: {str(state_mods)}",
            )
        )

    for ref in function_definition.get_all_references(False):
        if isinstance(ref, ir.IdentifierPathPart):
            top_statement = ref.underlying_node
        elif isinstance(ref, ir.ExternalReference):
            continue  # TODO currently not supported
        else:
            top_statement = ref
        func_call = None
        while top_statement is not None:
            if (
                func_call is None
                and isinstance(top_statement, ir.FunctionCall)
                and top_statement.function_called
                in get_all_base_and_child_functions(function_definition)
            ):
                func_call = top_statement
            if isinstance(top_statement, ir.StatementAbc):
                break
            top_statement = top_statement.parent

        if top_statement is None or func_call is None:
            continue
        function_def = top_statement
        while function_def is not None and not isinstance(
            function_def, ir.FunctionDefinition
        ):
            function_def = function_def.parent
        if function_def is None:
            continue
        assert isinstance(function_def, ir.FunctionDefinition)
        if top_statement in checked_statements:
            continue

        if source_address_declaration in function_definition.parameters.parameters:
            for arg_decl, arg_expr in pair_function_call_arguments(
                function_definition, func_call
            ):
                if arg_decl == source_address_declaration:
                    assert isinstance(arg_expr.type, (types.Address, types.Contract))
                    ret.extend(
                        check_reentrancy_in_function(
                            function_def,
                            top_statement,
                            arg_expr,
                            this_modifies_state,
                            checked_statements,
                        )
                    )
                    break
        else:
            ret.extend(
                check_reentrancy_in_function(
                    function_def,
                    top_statement,
                    address_source,
                    this_modifies_state,
                    checked_statements,
                )
            )
    return ret


class ReentrancyDetector(Detector):
    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_member_access(self, node: ir.MemberAccess):
        t = node.type
        if (
            not isinstance(t, types.Function)
            or t.kind
            not in {
                ir.enums.FunctionTypeKind.BARE_CALL,
                ir.enums.FunctionTypeKind.EXTERNAL,
            }
            or t.state_mutability
            in {ir.enums.StateMutability.PURE, ir.enums.StateMutability.VIEW}
        ):
            return

        address_source = node.expression

        function_call = node
        while function_call is not None:
            if isinstance(function_call, ir.FunctionCall):
                break
            function_call = function_call.parent
        if function_call is None:
            return

        if function_call.function_called != node.referenced_declaration:
            logger.debug(f"Re-entrancy ignored: {function_call.source}")
            return

        statement = function_call
        while statement is not None:
            if isinstance(statement, ir.StatementAbc):
                break
            statement = statement.parent
        if statement is None:
            return

        function_def = statement
        while function_def is not None:
            if isinstance(function_def, ir.FunctionDefinition):
                break
            function_def = function_def.parent
        if function_def is None:
            return

        ret = check_reentrancy_in_function(
            function_def, statement, address_source, set(), set()
        )
        if len(ret) == 0:
            return
        ret = list(
            sorted(ret, key=lambda x: (str(x.ir_node.file), x.ir_node.byte_location[0]))
        )

        self._detections.add(
            DetectorResult(
                Detection(
                    node,
                    f"Possible re-entrancy in `{function_def.canonical_name}`",
                    tuple(ret),
                ),
                impact=DetectionImpact.MEDIUM,
                confidence=DetectionConfidence.MEDIUM,
            )
        )

    @detector.command("reentrancy")
    def cli(self):
        """
        Detects re-entrancy vulnerabilities.
        """
        pass



import time
from contextlib import contextmanager, nullcontext
from typing import Any, Dict, Iterable, Optional, Union, cast
from urllib.error import HTTPError

import eth_utils
from rich.console import Group
from rich.pretty import pprint
from rich.progress import BarColumn, Progress, TextColumn, TimeElapsedColumn
from rich.prompt import Confirm
from rich.table import Table

import woke.development.core
from woke.cli.console import console
from woke.development.chain_interfaces import AnvilChainInterface, TxParams
from woke.development.core import (
    Abi,
    Address,
    RequestType,
    RevertToSnapshotFailedError,
    TransactionConfirmationFailedError,
    Wei,
    check_connected,
    fix_library_abi,
)
from woke.development.globals import chain_interfaces_manager, get_config
from woke.development.json_rpc.communicator import JsonRpcError
from woke.development.transactions import (
    Eip1559Transaction,
    Eip2930Transaction,
    LegacyTransaction,
    TransactionAbc,
    TransactionStatusEnum,
)
from woke.development.utils import chain_explorer_urls


class Chain(woke.development.core.Chain):
    @contextmanager
    def connect(
        self,
        uri: Optional[str] = None,
        *,
        accounts: Optional[int] = None,
        chain_id: Optional[int] = None,
        fork: Optional[str] = None,
        hardfork: Optional[str] = None,
        min_gas_price: Optional[Union[int, str]] = None,
        block_base_fee_per_gas: Optional[Union[int, str]] = None,
    ):
        yield from self._connect(
            uri,
            accounts=accounts,
            chain_id=chain_id,
            fork=fork,
            hardfork=hardfork,
            min_gas_price=min_gas_price,
            block_base_fee_per_gas=block_base_fee_per_gas,
        )

    def _connect_setup(
        self, min_gas_price: Optional[int], block_base_fee_per_gas: Optional[int]
    ) -> None:
        self._require_signed_txs = True

        if min_gas_price is not None:
            try:
                self._chain_interface.set_min_gas_price(min_gas_price)
            except (JsonRpcError, HTTPError):
                pass

    def _connect_finalize(self) -> None:
        chain_interfaces_manager.close(self._chain_interface)

    @check_connected
    def snapshot(self) -> str:
        snapshot_id = self._chain_interface.snapshot()

        self._snapshots[snapshot_id] = {
            "nonces": self._nonces.copy(),
            "accounts": self._accounts.copy(),
            "default_call_account": self._default_call_account,
            "default_tx_account": self._default_tx_account,
            "txs": dict(self._txs._transactions),
            "blocks": dict(self._blocks._blocks),
        }
        return snapshot_id

    @check_connected
    def revert(self, snapshot_id: str) -> None:
        reverted = self._chain_interface.revert(snapshot_id)
        if not reverted:
            raise RevertToSnapshotFailedError()

        snapshot = self._snapshots[snapshot_id]
        self._nonces = snapshot["nonces"]
        self._accounts = snapshot["accounts"]
        self._default_call_account = snapshot["default_call_account"]
        self._default_tx_account = snapshot["default_tx_account"]
        self._txs._transactions = snapshot["txs"]
        self._blocks._blocks = snapshot["blocks"]
        del self._snapshots[snapshot_id]

    @property
    @check_connected
    def block_gas_limit(self) -> int:
        return self._blocks["pending"].gas_limit

    @block_gas_limit.setter
    @check_connected
    def block_gas_limit(self, value: int) -> None:
        raise NotImplementedError(
            "Cannot set block gas limit in deployment"
        )  # TODO do nothing instead?

    @property
    @check_connected
    def gas_price(self) -> Wei:
        return Wei(self.chain_interface.get_gas_price())

    @gas_price.setter
    @check_connected
    def gas_price(self, value: int) -> None:
        raise NotImplementedError(
            "Cannot set gas price in deployment"
        )  # TODO do nothing instead?

    @property
    @check_connected
    def max_priority_fee_per_gas(self) -> Wei:
        return Wei(self.chain_interface.get_max_priority_fee_per_gas())

    @max_priority_fee_per_gas.setter
    @check_connected
    def max_priority_fee_per_gas(self, value: int) -> None:
        raise NotImplementedError(
            "Cannot set max priority fee per gas in deployment"
        )  # TODO do nothing instead?

    def _build_transaction(
        self,
        request_type: RequestType,
        params: TxParams,
        arguments: Iterable,
        abi: Optional[Dict],
    ) -> TxParams:
        tx_type = params.get("type", self._default_tx_type)
        if tx_type not in {0, 1, 2}:
            raise ValueError("Invalid transaction type")

        if tx_type == 0 and (
            "accessList" in params
            or "maxFeePerGas" in params
            or "maxPriorityFeePerGas" in params
        ):
            raise ValueError(
                "Cannot specify accessList, maxFeePerGas, or maxPriorityFeePerGas for type 0 transaction"
            )
        elif tx_type == 1 and (
            "maxFeePerGas" in params or "maxPriorityFeePerGas" in params
        ):
            raise ValueError(
                "Cannot specify maxFeePerGas or maxPriorityFeePerGas for type 1 transaction"
            )
        elif tx_type == 2 and "gasPrice" in params:
            raise ValueError("Cannot specify gasPrice for type 2 transaction")

        if "from" in params:
            sender = params["from"]
        else:
            if request_type == "call" and self.default_call_account is not None:
                sender = str(self.default_call_account.address)
            elif request_type == "tx" and self.default_tx_account is not None:
                sender = str(self.default_tx_account.address)
            elif (
                request_type == "estimate" and self.default_estimate_account is not None
            ):
                sender = str(self.default_estimate_account.address)
            elif (
                request_type == "access_list"
                and self.default_access_list_account is not None
            ):
                sender = str(self.default_access_list_account.address)
            else:
                raise ValueError(
                    "No from_ account specified and no default account set"
                )

        if "data" not in params:
            params["data"] = b""

        if abi is None:
            params["data"] += Abi.encode([], [])
        else:
            arguments = [self._convert_to_web3_type(arg) for arg in arguments]
            types = [
                eth_utils.abi.collapse_if_tuple(cast(Dict[str, Any], arg))
                for arg in fix_library_abi(abi["inputs"])
            ]
            params["data"] += Abi.encode(types, arguments)

        n = self._nonces[Address(sender)]
        tx: TxParams = {
            "nonce": n,
            "from": sender,
            "value": params["value"] if "value" in params else 0,
            "data": params["data"],
        }
        if tx_type != 0:
            tx["type"] = tx_type

        if "to" in params:
            tx["to"] = params["to"]

        if tx_type == 0:
            tx["gasPrice"] = (
                params["gasPrice"] if "gasPrice" in params else self.gas_price
            )
        elif tx_type == 1:
            tx["chainId"] = self._chain_id
            tx["gasPrice"] = (
                params["gasPrice"] if "gasPrice" in params else self.gas_price
            )
        elif tx_type == 2:
            tx["chainId"] = self._chain_id
            tx["maxPriorityFeePerGas"] = (
                params["maxPriorityFeePerGas"]
                if "maxPriorityFeePerGas" in params
                else self.max_priority_fee_per_gas
            )
            if "maxFeePerGas" in params:
                tx["maxFeePerGas"] = params["maxFeePerGas"]
            else:
                if (
                    isinstance(self.chain_interface, AnvilChainInterface)
                    or self.require_signed_txs
                ):
                    tx["maxFeePerGas"] = tx["maxPriorityFeePerGas"] + int(
                        int(
                            self.chain_interface.get_block("pending")["baseFeePerGas"],
                            16,
                        )
                        * 2
                    )

        if "gas" not in params or params["gas"] == "auto":
            # use "auto when unset
            try:
                tx_copy = tx.copy()
                tx_copy.pop("gasPrice", None)
                tx_copy.pop("maxPriorityFeePerGas", None)
                tx_copy.pop("maxFeePerGas", None)
                tx["gas"] = int(self._chain_interface.estimate_gas(tx_copy) * 1.1)
            except JsonRpcError as e:
                self._process_call_revert(e)
                raise
        elif isinstance(params["gas"], int):
            tx["gas"] = params["gas"]
        else:
            raise ValueError(f"Invalid gas value: {params['gas']}")

        if (
            tx_type in {1, 2}
            and ("accessList" not in params or params["accessList"] == "auto")
            and request_type != "access_list"
        ):
            try:
                response = self._chain_interface.create_access_list(tx)
                gas_used = int(int(response["gasUsed"], 16) * 1.1)

                if params.get("accessList", None) == "auto" or (
                    "accessList" not in params and gas_used <= tx["gas"]
                ):
                    tx["accessList"] = response["accessList"]

                    if "gas" not in params or params["gas"] == "auto":
                        tx["gas"] = gas_used
            except (JsonRpcError, HTTPError) as e:
                try:
                    if isinstance(e, JsonRpcError):
                        # will re-raise if not a revert error
                        self._process_call_revert(e)
                        raise
                    else:
                        # HTTPError -> eth_createAccessList not supported
                        if "accessList" not in params:
                            tx["accessList"] = []
                        else:
                            raise
                except JsonRpcError:
                    # eth_createAccessList probably not supported
                    if "accessList" not in params:
                        tx["accessList"] = []
                    else:
                        raise

        return tx

    def _wait_for_transaction(
        self, tx: TransactionAbc, confirmations: Optional[int]
    ) -> None:
        def get_pending_text():
            if tx.chain.chain_id in chain_explorer_urls:
                text = f"Waiting for transaction [link={chain_explorer_urls[tx.chain.chain_id].url}/tx/{tx.tx_hash}]{tx.tx_hash}[/link] to be mined\n"
            else:
                text = f"Waiting for transaction {tx.tx_hash} to be mined\n"

            t = Table("", "Set in transaction", "Current recommended")

            if isinstance(tx, Eip1559Transaction):
                recommended_priority_fee = (
                    self.chain_interface.get_max_priority_fee_per_gas() / 10**9
                )
                base_fee = (
                    int(self.chain_interface.get_block("pending")["baseFeePerGas"], 16)
                    / 10**9
                )

                t.add_row(
                    "Max fee per gas",
                    f"{(tx.max_fee_per_gas / 10 ** 9):.2f} Gwei",
                    f"{(recommended_priority_fee + base_fee):.2f} Gwei",
                )
                t.add_row(
                    "Max priority fee per gas",
                    f"{(tx.max_priority_fee_per_gas / 10 ** 9):.2f} Gwei",
                    f"{(recommended_priority_fee):.2f} Gwei",
                )
            elif isinstance(tx, (Eip2930Transaction, LegacyTransaction)):
                t.add_row(
                    "Gas price",
                    f"{(tx.gas_price / 10 ** 9):.2f} Gwei",
                    f"{(self.chain_interface.get_gas_price() / 10 ** 9):.2f} Gwei",
                )

            return Group(text, t)

        if confirmations == 0:
            return
        elif confirmations is None:
            confirmations = self.default_tx_confirmations

        config = get_config()

        ctx_manager = (
            console.status(get_pending_text())
            if not config.deployment.silent
            else nullcontext()
        )

        with ctx_manager as status:
            while tx.status == TransactionStatusEnum.PENDING:
                time.sleep(0.5)
                if status is not None:
                    status.update(get_pending_text())

        if not config.deployment.silent:
            if tx.chain.chain_id in chain_explorer_urls:
                console.print(
                    f"Transaction [link={chain_explorer_urls[tx.chain.chain_id].url}/tx/{tx.tx_hash}]{tx.tx_hash}[/link] mined in block {tx.block_number}"
                )
            else:
                console.print(
                    f"Transaction {tx.tx_hash} mined in block {tx.block_number}"
                )

        latest_block_number = self.chain_interface.get_block_number()

        ctx_manager = (
            Progress(
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                TextColumn("{task.completed} of {task.total}"),
                TimeElapsedColumn(),
                console=console,
            )
            if not config.deployment.silent
            else nullcontext()
        )

        with ctx_manager as progress:
            if progress is not None:
                task_id = progress.add_task(
                    "Confirmations",
                    total=confirmations,
                    completed=(latest_block_number - tx.block_number + 1),
                )
            while latest_block_number - tx.block_number < confirmations - 1:
                time.sleep(1)
                latest_block_number = self.chain_interface.get_block_number()
                if progress is not None:
                    progress.update(
                        task_id,  # pyright: ignore reportUnboundVariable
                        completed=(latest_block_number - tx.block_number + 1),
                    )

    def _confirm_transaction(self, tx: TxParams) -> None:
        config = get_config()
        if config.deployment.silent:
            return
        pprint(tx, console=console, max_string=200)

        if config.deployment.confirm_transactions:
            confirm = Confirm.ask("Sign and send transaction?")
            if not confirm:
                raise TransactionConfirmationFailedError()


default_chain = Chain()



from __future__ import annotations

import pathlib
import shutil
from os import PathLike
from typing import Union


def is_relative_to(path: pathlib.PurePath, *other: Union[str, PathLike[str]]):
    """
    Return True if the path is relative to another path or False.
    Backported from Python 3.9 (https://github.com/python/cpython/blob/75a6441718dcbc65d993c9544e67e25bef120e82/Lib/pathlib.py#L687-L694)
    """
    try:
        path.relative_to(*other)
        return True
    except ValueError:
        return False


def copy_dir(
    src_dir: pathlib.Path,
    dst_dir: pathlib.Path,
    *,
    overwrite: bool = False,
    raise_on_existing: bool = False,
) -> None:
    """
    Copy contents of src_dir and create dst_dir if it doesn't exist.
    Overwrite files if overwrite is True. Raise FileExistsError if raise_on_existing is True and
    existing files are found.
    """
    dst_dir.mkdir(exist_ok=True)
    src_dst_paths = [
        (p.absolute(), dst_dir / p.relative_to(src_dir)) for p in src_dir.rglob("*")
    ]
    if not overwrite and raise_on_existing:
        existing_files = []
        for _, dst_path in src_dst_paths:
            if dst_path.exists():
                existing_files.append(str(dst_path))
        if existing_files:
            raise FileExistsError(f"Existing files: {existing_files}")

    for src_path, dst_path in src_dst_paths:
        if src_path.is_dir():
            dst_path.mkdir(parents=True, exist_ok=True)
        else:
            if not dst_path.parent.exists():
                dst_path.parent.mkdir(parents=True, exist_ok=True)
            if overwrite or not dst_path.exists():
                shutil.copy(str(src_path), str(dst_path))



from __future__ import annotations

import re
from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Tuple, Union

from ..type_names.elementary_type_name import ElementaryTypeName
from .abc import DeclarationAbc

if TYPE_CHECKING:
    from .contract_definition import ContractDefinition
    from ..meta.source_unit import SourceUnit

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcUserDefinedValueTypeDefinition
from woke.ir.utils import IrInitTuple


class UserDefinedValueTypeDefinition(DeclarationAbc):
    """
    Definition of a user defined value type.

    !!! example
        ```solidity
        type MyInt is uint;
        ```
    """

    _ast_node: SolcUserDefinedValueTypeDefinition
    _parent: Union[ContractDefinition, SourceUnit]

    _underlying_type: ElementaryTypeName

    def __init__(
        self,
        init: IrInitTuple,
        user_defined_value_type_definition: SolcUserDefinedValueTypeDefinition,
        parent: SolidityAbc,
    ):
        super().__init__(init, user_defined_value_type_definition, parent)
        self._underlying_type = ElementaryTypeName(
            init, user_defined_value_type_definition.underlying_type, self
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._underlying_type

    def _parse_name_location(self) -> Tuple[int, int]:
        IDENTIFIER = r"[a-zA-Z$_][a-zA-Z0-9$_]*"
        USER_DEF_VAL_TYPE_RE = re.compile(
            r"^\s*type\s+(?P<name>{identifier})".format(identifier=IDENTIFIER).encode(
                "utf-8"
            )
        )

        byte_start = self._ast_node.src.byte_offset
        match = USER_DEF_VAL_TYPE_RE.match(self._source)
        assert match
        return byte_start + match.start("name"), byte_start + match.end("name")

    @property
    def parent(self) -> Union[ContractDefinition, SourceUnit]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    @lru_cache(maxsize=2048)
    def canonical_name(self) -> str:
        from .contract_definition import ContractDefinition

        if isinstance(self._parent, ContractDefinition):
            return f"{self._parent.canonical_name}.{self._name}"
        return self._name

    @property
    @lru_cache(maxsize=2048)
    def declaration_string(self) -> str:
        return f"type {self.name} is {self._underlying_type.source}"

    @property
    def underlying_type(self) -> ElementaryTypeName:
        """
        Returns:
            Underlying type of the user defined value type.
        """
        return self._underlying_type



from typing import Any, List, Optional, Union

from woke.lsp.common_structures import (
    ExecuteCommandOptions,
    PositionEncodingKind,
    WorkspaceSymbolOptions,
)
from woke.lsp.document_sync import TextDocumentSyncKind, TextDocumentSyncOptions
from woke.utils import StrEnum

from .features.code_action import CodeActionOptions
from .features.code_lens import CodeLensOptions
from .features.completion import CompletionOptions
from .features.definition import DefinitionOptions
from .features.document_link import DocumentLinkOptions
from .features.document_symbol import DocumentSymbolOptions
from .features.hover import HoverOptions
from .features.implementation import (
    ImplementationOptions,
    ImplementationRegistrationOptions,
)
from .features.references import ReferenceOptions
from .features.rename import RenameOptions
from .features.type_definition import (
    TypeDefinitionOptions,
    TypeDefinitionRegistrationOptions,
)
from .features.type_hierarchy import (
    TypeHierarchyOptions,
    TypeHierarchyRegistrationOptions,
)
from .lsp_data_model import LspModel


class WorkspaceFoldersServerCapabilities(LspModel):
    supported: Optional[bool] = None
    change_notifications: Optional[Union[str, bool]] = None


class FileOperationPatternKind(StrEnum):
    FILE = "file"
    FOLDER = "folder"


class FileOperationPatternOptions(LspModel):
    ignore_case: Optional[bool]


class FileOperationPattern(LspModel):
    glob: str
    """
    The glob pattern to match. Glob patterns can have the following syntax:
    - `*` to match one or more characters in a path segment
    - `?` to match one character in a path segment
    - `**` to match one or more characters in a path segment, including none
    - `{}` to group sub patterns into an OR expression. (e.g. `**/*.{ts,js}`
        matches all TypeScript and JavaScript files)
    - `[]` to declare a range of characters to match in a path segment
        (e.g., `example.[0-9]` to match `example.0`, `example.1`, …)
    - `[!...]` to negate a range of characters to match in a path segment
        (e.g., `example.[!0-9]` to match `example.a`, `example.b`, but
        not `example.0`)
    """
    matches: Optional[FileOperationPatternKind] = None
    """
    Whether to match files or folders with this pattern.

    Matches both if undefined.
    """
    options: Optional[FileOperationPatternOptions] = None
    """
    Additional options used during matching.
    """


class FileOperationFilter(LspModel):
    scheme: Optional[str] = None
    """
    A Uri like `file` or `untitled`.
    """
    pattern: FileOperationPattern
    """
    The actual file operation pattern.
    """


class FileOperationRegistrationOptions(LspModel):
    filters: List[FileOperationFilter]
    """
    The actual filters.
    """


class ServerCapabilitiesWorkspaceFileOperations(LspModel):
    """
    ServerCapabilities subsubClass
    """

    did_create: Optional[FileOperationRegistrationOptions] = None
    will_create: Optional[FileOperationRegistrationOptions] = None
    did_rename: Optional[FileOperationRegistrationOptions] = None
    will_rename: Optional[FileOperationRegistrationOptions] = None
    did_delete: Optional[FileOperationRegistrationOptions] = None
    will_delete: Optional[FileOperationRegistrationOptions] = None


class ServerCapabilitiesWorkspace(LspModel):
    """
    ServerCapabilities subClass
    """

    workspace_folders: Optional[WorkspaceFoldersServerCapabilities] = None
    file_operations: Optional[ServerCapabilitiesWorkspaceFileOperations] = None


class ServerCapabilities(LspModel):
    position_encoding: Optional[PositionEncodingKind] = None
    text_document_sync: Optional[
        Union[TextDocumentSyncOptions, TextDocumentSyncKind]
    ] = None
    document_link_provider: Optional[DocumentLinkOptions] = None
    type_hierarchy_provider: Optional[
        Union[bool, TypeHierarchyOptions, TypeHierarchyRegistrationOptions]
    ] = None
    references_provider: Optional[Union[bool, ReferenceOptions]] = None
    document_symbol_provider: Optional[Union[bool, DocumentSymbolOptions]] = None
    definition_provider: Optional[Union[bool, DefinitionOptions]] = None
    implementation_provider: Optional[
        Union[bool, ImplementationOptions, ImplementationRegistrationOptions]
    ] = None
    type_definition_provider: Optional[
        Union[bool, TypeDefinitionOptions, TypeDefinitionRegistrationOptions]
    ] = None
    code_lens_provider: Optional[CodeLensOptions] = None
    rename_provider: Optional[Union[bool, RenameOptions]] = None
    execute_command_provider: Optional[ExecuteCommandOptions] = None
    hover_provider: Optional[Union[bool, HoverOptions]] = None
    completion_provider: Optional[CompletionOptions] = None
    code_action_provider: Optional[Union[bool, CodeActionOptions]] = None
    """
    signature_help_provider: Optional[SignatureHelpOptions]
    declaration_provider: Optional[Union[bool, DeclarationOptions, DeclarationRegistrationOptions]]
    document_highlight_provider: Optional[Union[bool, ReferenceOptions]]
    color_provider: Optional[Union[bool, DocumentColorOptions, DocumentColorRegistrationOptions]]
    document_formatting_provider: Optional[Union[bool, DocumentFormattingOptions]]
    document_range_formatting_provider: Optional[Union[bool, DocumentRangeFormattingOptions]]
    document_on_type_formatting_provider: Optional[DocumentOnTypeFormattingOptions]
    folding_range_provider: Optional[Union[bool, FoldingRangeOptions, FoldingRangeRegistrationOptions]]
    selection_range_provider: Optional[Union[bool, SelectionRangeOptions, SelectionRangeRegistrationOptions]]
    linked_editing_range_provider: Optional[Union[bool, LinkedEditingRangeOptions, LinkedEditingRangeRegistrationOptions]]
    call_hierarchy_provider: Optional[Union[bool, CallHierarchyOptions, CallHierarchyRegistrationOptions]]
    semantic_token_provider: Optional[Union[SemanticTokensOptions, SemanticTokensRegistrationOptions]]
    moniker_provider: Optional[Union[bool, MonikerOptions, MonikerRegistrationOptions]]
    diagnostic_provider: Optional[
        Union[DiagnosticOptions, DiagnosticRegistrationOptions]
    ] = None
    """
    workspace_symbol_provider: Optional[Union[bool, WorkspaceSymbolOptions]] = None
    workspace: Optional[ServerCapabilitiesWorkspace] = None
    experimental: Optional[Any] = None


class InitializeResultServerInfo(LspModel):
    name: str
    version: Optional[str]


class InitializeResult(LspModel):
    capabilities: ServerCapabilities
    server_info: Optional[InitializeResultServerInfo]



import logging
from pathlib import Path
from typing import List, Optional, Tuple, Union

from woke.ir import (
    BinaryOperation,
    ContractDefinition,
    DeclarationAbc,
    Identifier,
    IdentifierPath,
    IrAbc,
    MemberAccess,
    UnaryOperation,
    UserDefinedTypeName,
    YulIdentifier,
)

from ...core.solidity_version import SemanticVersion
from ..common_structures import (
    MarkupContent,
    MarkupKind,
    Position,
    Range,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from ..context import LspContext
from ..lsp_data_model import LspModel
from ..utils import position_within_range, uri_to_path
from ..utils.position import changes_to_byte_offset

logger = logging.getLogger(__name__)


class HoverClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    content_format: Optional[List[MarkupKind]]


class HoverOptions(WorkDoneProgressOptions):
    pass


class HoverRegistrationOptions(TextDocumentRegistrationOptions, HoverOptions):
    pass


class HoverParams(TextDocumentPositionParams, WorkDoneProgressParams):
    pass


class MarkedString(LspModel):
    language: str
    value: str


MarkedStringType = Union[str, MarkedString]


class Hover(LspModel):
    contents: Union[MarkedStringType, List[MarkedStringType], MarkupContent]
    range: Optional[Range]


def _append_openzeppelin_docs(
    node: ContractDefinition, version: SemanticVersion
) -> str:
    source_unit = node.parent
    version_string = f"{version.major}.x"
    url_base = f"https://docs.openzeppelin.com/contracts/{version_string}"
    doc_url = None
    api_doc_url = None

    if "openzeppelin/contracts/access" in source_unit.source_unit_name:
        doc_url = "access-control"
        api_doc_url = "access"
    elif "openzeppelin/contracts/crosschain" in source_unit.source_unit_name:
        doc_url = "crosschain"
        api_doc_url = "crosschain"
    elif "openzeppelin/contracts/finance" in source_unit.source_unit_name:
        if node.name == "PaymentSplitter":
            doc_url = "utilities#payment"
        api_doc_url = "finance"
    elif "openzeppelin/contracts/governance" in source_unit.source_unit_name:
        doc_url = "governance"
        api_doc_url = "governance"
    elif (
        "openzeppelin/contracts/interfaces" in source_unit.source_unit_name
        or "openzeppelin/contracts/token" in source_unit.source_unit_name
    ):
        if "ERC20" in node.name:
            doc_url = "erc20"
            api_doc_url = "token/erc20"
        elif "ERC721" in node.name:
            doc_url = "erc721"
            api_doc_url = "token/erc721"
        elif "ERC777" in node.name:
            doc_url = "erc777"
            api_doc_url = "token/erc777"
        elif "ERC1155" in node.name:
            doc_url = "erc1155"
            api_doc_url = "token/erc1155"
        elif node.name.startswith("I"):
            api_doc_url = "interfaces"
        else:
            api_doc_url = "token/common"
    elif "openzeppelin/contracts/metatx" in source_unit.source_unit_name:
        api_doc_url = "metatx"
    elif "openzeppelin/contracts/proxy" in source_unit.source_unit_name:
        api_doc_url = "proxy"
    elif "openzeppelin/contracts/security" in source_unit.source_unit_name:
        if node.name == "PullPayment":
            doc_url = "utilities#payment"
        api_doc_url = "security"
    elif "openzeppelin/contracts/utils" in source_unit.source_unit_name:
        doc_url = "utilities"
        api_doc_url = "utils"
    elif "openzeppelin/contracts/GSN" in source_unit.source_unit_name:
        doc_url = "gsn"
        api_doc_url = "gsn"
    elif "openzeppelin/contracts/cryptography" in source_unit.source_unit_name:
        doc_url = "utilities#cryptography"
        api_doc_url = "cryptography"
    elif "openzeppelin/contracts/drafts" in source_unit.source_unit_name:
        api_doc_url = "drafts"
    elif "openzeppelin/contracts/math" in source_unit.source_unit_name:
        doc_url = "utilities#math"
        api_doc_url = "math"
    elif "openzeppelin/contracts/payment" in source_unit.source_unit_name:
        doc_url = "utilities#payment"
        api_doc_url = "payment"
    elif "openzeppelin/contracts/presets" in source_unit.source_unit_name:
        if "ERC20" in node.name:
            doc_url = "erc20#Presets"
        elif "ERC721" in node.name:
            doc_url = "erc721#Presets"
        elif "ERC777" in node.name:
            doc_url = "erc777#Presets"
        elif "ERC1155" in node.name:
            doc_url = "erc1155#Presets"
        api_doc_url = "presets"

    ret = ""
    if doc_url is not None:
        ret += f"\n\n[OpenZeppelin documentation]({url_base}/{doc_url})"
    if api_doc_url is not None:
        ret += f"\n\n[OpenZeppelin API documentation]({url_base}/api/{api_doc_url}#{node.name})"
    return ret


yul_definitions = {
    "stop": "// stop execution, identical to return(0, 0)\nstop()",
    "add": "// x + y\nadd(x, y)",
    "sub": "// x - y\nsub(x, y)",
    "mul": "// x * y\nmul(x, y)",
    "div": "// x / y, 0 if y == 0\ndiv(x, y)",
    "sdiv": "// signed x / y, 0 if y == 0\nsdiv(x, y)",
    "mod": "// x % y, 0 if y == 0\nmod(x, y)",
    "smod": "// signed x % y, 0 if y == 0\nsmod(x, y)",
    "exp": "// x ** y\nexp(x, y)",
    "not": "// bitwise NOT\nnot(x)",
    "lt": "// 1 if x < y, 0 otherwise\nlt(x, y)",
    "gt": "// 1 if x > y, 0 otherwise\ngt(x, y)",
    "slt": "// 1 if signed x < signed y, 0 otherwise\nslt(x, y)",
    "sgt": "// 1 if signed x > signed y, 0 otherwise\nsgt(x, y)",
    "eq": "// 1 if x == y, 0 otherwise\neq(x, y)",
    "iszero": "// 1 if x == 0, 0 otherwise\niszero(x)",
    "and": "// bitwise AND\nand(x, y)",
    "or": "// bitwise OR\nor(x, y)",
    "xor": "// bitwise XOR\nxor(x, y)",
    "byte": "// retrieve the nth most significant byte of x\nbyte(n, x)",
    "shl": "// x << y\nshl(x, y)",
    "shr": "// x >> y\nshr(x, y)",
    "sar": "// signed x >> y\nsar(x, y)",
    "addmod": "// (x + y) % m, 0 if m == 0\naddmod(x, y, m)",
    "mulmod": "// (x * y) % m, 0 if m == 0\nmulmod(x, y, m)",
    "signextend": "// sign extend from (i * 8 + 7)th bit counting from least significant\nsignextend(i, x)",
    "keccak256": "// compute the Keccak-256 hash\nkeccak256(memStart, length)",
    "pc": "// current program counter\npc()",
    "pop": "// discard value returned by another instruction\npop(x)",
    "mload": "// load word from memory\nmload(memStart)",
    "mstore": "// store word to memory\nmstore(memStart, x)",
    "mstore8": "// store single byte to memory\nmstore8(memStart, x)",
    "sload": "// load word from storage\nsload(storageSlot)",
    "sstore": "// store word to storage\nsstore(storageSlot, x)",
    "msize": "// size of memory\nmsize()",
    "gas": "// remaining gas\ngas()",
    "address": "// address of the current contract / execution context\naddress()",
    "balance": "// wei balance of an address\nbalance(address)",
    "selfbalance": "// wei balance of the current contract / execution context, same as balance(address()) but cheaper\nselfbalance()",
    "caller": "// caller (excluding delegatecall), same as `msg.sender` in Solidity\ncaller()",
    "callvalue": "// wei sent with the call, same as `msg.value` in Solidity\ncallvalue()",
    "calldataload": "// load word from call data\ncalldataload(callDataStart)",
    "calldatasize": "// size of call data in bytes\ncalldatasize()",
    "calldatacopy": "// copy call data to memory\ncalldatacopy(memStart, callDataStart, length)",
    "codesize": "// size of code of the current contract / execution context\ncodesize()",
    "codecopy": "// copy code to memory\ncodecopy(memStart, codeStart, length)",
    "extcodesize": "// size of code at address\nextcodesize(address)",
    "extcodecopy": "// copy code at address to memory\nextcodecopy(address, memStart, codeStart, length)",
    "returndatasize": "// size of return data buffer\nreturndatasize()",
    "returndatacopy": "// copy return data to memory\nreturndatacopy(memStart, returnDataStart, length)",
    "extcodehash": "// Keccak-256 hash of code at address\nextcodehash(address)",
    "create": "// create contract with code from memory and value in wei, return new address or 0 on error\ncreate(value, memStart, length)",
    "create2": "// create contract from salt with code from memory and value in wei, return new address or 0 on error\ncreate2(value, memStart, length, salt)",
    "call": "// call contract at address providing gas, value in wei, call data from memory, storing return data to memory, return 1 on success or 0 on error\ncall(gas, address, value, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "callcode": "// call contract at address preserving execution context, providing gas, value in wei, call data from memory, storing return data to memory, return 1 on success or 0 on error\ncallcode(gas, address, value, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "delegatecall": "// call contract at address preserving execution context, current caller and call value, providing gas, call data from memory, storing return data to memory, return 1 on success or 0 on error\ndelegatecall(gas, address, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "staticcall": "// call contract at address disallowing state modifications, providing gas, call data from memory, storing return data to memory, return 1 on success or 0 on error\nstaticcall(gas, address, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "return": "// end execution, copy memory to return data buffer\nreturn(memStart, length)",
    "revert": "// end execution, revert state changes, copy memory to return data buffer\nrevert(memStart, length)",
    "selfdestruct": "// end execution, mark current contract to be destroyed, sending funds to address\nselfdestruct(address)",
    "invalid": "// end execution with invalid instruction, consume all remaining gas\ninvalid()",
    "log0": "// log without topics data from memory\nlog0(memStart, length)",
    "log1": "// log with one topic data from memory\nlog1(memStart, length, topic1)",
    "log2": "// log with two topics data from memory\nlog2(memStart, length, topic1, topic2)",
    "log3": "// log with three topics data from memory\nlog3(memStart, length, topic1, topic2, topic3)",
    "log4": "// log with four topics data from memory\nlog4(memStart, length, topic1, topic2, topic3, topic4)",
    "chainid": "// current chain ID\nchainid()",
    "basefee": "// current block's base fee\nbasefee()",
    "origin": "// sender of the transaction, same as `tx.origin` in Solidity\norigin()",
    "gasprice": "// gas price of the transaction, same as `tx.gasprice` in Solidity\ngasprice()",
    "blockhash": "// hash of the given block number, only for last 256 blocks excluding current\nblockhash(blockNumber)",
    "coinbase": "// current block's miner address\ncoinbase()",
    "timestamp": "// timestamp of the current block in seconds since the epoch\ntimestamp()",
    "number": "// current block number\nnumber()",
    "difficulty": "// current block's difficulty\ndifficulty()",
    "gaslimit": "// current block's gas limit\ngaslimit()",
    "prevrandao": "// previous block's RANDAO value\nprevrandao()",
}


def _get_results_from_node(
    original_node: IrAbc,
    position: Position,
    byte_offset: int,
    context: LspContext,
    node_name_location: Optional[Tuple[int, int]],
) -> Optional[Tuple[str, Tuple[int, int]]]:
    original_node_location: Tuple[int, int] = original_node.byte_location
    logger.debug(f"Found node {original_node}")

    if isinstance(original_node, DeclarationAbc):
        assert node_name_location is not None
        name_location_range = context.compiler.get_range_from_byte_offsets(
            original_node.file, node_name_location
        )
        if not position_within_range(position, name_location_range):
            return None

    if isinstance(original_node, (Identifier, MemberAccess)):
        node = original_node.referenced_declaration
        if node is None:
            return None
    elif isinstance(original_node, (IdentifierPath, UserDefinedTypeName)):
        part = original_node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
        original_node_location = part.byte_location
    elif isinstance(original_node, YulIdentifier):
        external_reference = original_node.external_reference
        if external_reference is not None:
            node = external_reference.referenced_declaration
            original_node_location = external_reference.byte_location
        else:
            node = original_node
    elif (
        isinstance(original_node, (UnaryOperation, BinaryOperation))
        and original_node.function is not None
    ):
        node = original_node.function
    else:
        node = original_node

    if isinstance(node, DeclarationAbc):
        value = "```solidity\n" + node.declaration_string + "\n```"
        if (
            isinstance(node, ContractDefinition)
            and context.openzeppelin_contracts_version is not None
            and context.openzeppelin_contracts_version >= "2.0.0"
        ):
            value += _append_openzeppelin_docs(
                node, context.openzeppelin_contracts_version
            )

        return value, original_node_location
    elif isinstance(node, YulIdentifier):
        if node.name in yul_definitions:
            return (
                f"```solidity\n{yul_definitions[node.name]}\n```",
                original_node_location,
            )
    return None


async def _get_hover_from_cache(path: Path, position: Position, context: LspContext):
    new_byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, position.line, position.character
    )
    backward_changes = context.compiler.get_last_compilation_backward_changes(path)
    if backward_changes is None:
        raise Exception("No backward changes found")
    changes_before = backward_changes[0:new_byte_offset]
    old_byte_offset = changes_to_byte_offset(changes_before) + new_byte_offset

    tree = context.compiler.last_compilation_interval_trees[path]

    intervals = tree.at(old_byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    if len(nodes) == 0:
        raise ValueError(f"Could not find node at {old_byte_offset}")

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, DeclarationAbc):
        location = node.name_location
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        node_name_location = (new_start, new_end)
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, position, old_byte_offset, context, node_name_location
    )
    if result is None:
        return None

    value, location = result
    forward_changes = context.compiler.get_last_compilation_forward_changes(path)
    if forward_changes is None:
        raise Exception("No forward changes found")
    new_start = changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
    new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
    return Hover(
        contents=MarkupContent(
            kind=MarkupKind.MARKDOWN,
            value=value,
        ),
        range=context.compiler.get_range_from_byte_offsets(path, (new_start, new_end)),
    )


async def hover(context: LspContext, params: HoverParams) -> Optional[Hover]:
    logger.debug(
        f"Hover for file {params.text_document.uri} at position {params.position} requested"
    )

    path = uri_to_path(params.text_document.uri).resolve()
    if (
        path not in context.compiler.interval_trees
        or not context.compiler.output_ready.is_set()
    ):
        try:
            return await _get_hover_from_cache(path, params.position, context)
        except Exception:
            pass

    await context.compiler.output_ready.wait()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(f"Found {len(nodes)} nodes at position {params.position}")
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, DeclarationAbc):
        node_name_location = node.name_location
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, params.position, byte_offset, context, node_name_location
    )
    if result is None:
        return None

    value, original_node_location = result
    return Hover(
        contents=MarkupContent(
            kind=MarkupKind.MARKDOWN,
            value=value,
        ),
        range=context.compiler.get_range_from_byte_offsets(
            path, original_node_location
        ),
    )



from __future__ import annotations

import logging
from typing import List, Set

import woke.ir as ir
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)

logger = logging.getLogger(__name__)


class UnsafeSelfdestructDetector(Detector):
    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_call(self, node: ir.FunctionCall):
        from woke.analysis.ownable import statement_is_publicly_executable

        if node.function_called not in {
            ir.enums.GlobalSymbolsEnum.SELFDESTRUCT,
            ir.enums.GlobalSymbolsEnum.SUICIDE,
        }:
            return

        current_node = node
        while current_node is not None:
            if isinstance(current_node, ir.StatementAbc):
                break
            current_node = current_node.parent
        if current_node is None:
            return
        if not statement_is_publicly_executable(current_node):
            return

        self._detections.add(
            DetectorResult(
                Detection(node, "Selfdestruct call is not protected"),
                impact=DetectionImpact.HIGH,
                confidence=DetectionConfidence.MEDIUM,
            )
        )

    @detector.command("unsafe-selfdestruct")
    def cli(self):
        """
        Selfdestruct call is not protected.
        """
        pass



import re


class SolidityImportExpr:
    """
    Solidity import expression parser. Correctness of a whole expression is checked, but only filename is extracted.
    At this stage, filename can be any Solidity string. No further checks are performed.
    """

    FILENAME = r"""(?P<filename>'.*[^\\]'|".*[^\\]")"""
    SYMBOL = r"[_a-zA-Z][_a-zA-Z0-9]*"
    ALIAS = r"\s*{symbol}(?:\s+as\s+{symbol})?\s*".format(symbol=SYMBOL)
    IMPORT_FILENAME_RE = re.compile(r"\s*{filename}\s*".format(filename=FILENAME))
    IMPORT_AS_FROM_RE = re.compile(
        r"\s*\*\s*as\s+{symbol}\s+from\s*{filename}\s*".format(
            filename=FILENAME, symbol=SYMBOL
        )
    )
    IMPORT_AS_RE = re.compile(
        r"\s*{filename}\s*as\s+{symbol}\s*".format(filename=FILENAME, symbol=SYMBOL)
    )
    IMPORT_ALIAS_LIST = re.compile(
        r"\s*{{{alias}(?:,{alias})*}}\s*from\s*{filename}\s*".format(
            alias=ALIAS, filename=FILENAME
        )
    )

    __filename: str

    def __init__(self, expr: str):
        cls = self.__class__
        res = (
            cls.IMPORT_FILENAME_RE,
            cls.IMPORT_AS_FROM_RE,
            cls.IMPORT_AS_RE,
            cls.IMPORT_ALIAS_LIST,
        )
        matches = list(re.match(expr) for re in res)

        if not any(matches):
            raise ValueError(f"Invalid import expression: `{expr}`")

        match = next(match for match in matches if match is not None)
        # strip leading and trailing quote and replace escapes
        self.__filename = (
            match.groupdict()["filename"][1:-1].replace("\\'", "'").replace('\\"', '"')
        )

    @property
    def filename(self):
        return self.__filename



from functools import lru_cache
from typing import Iterator, Optional, Set, Tuple

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcIndexRangeAccess
from woke.ir.enums import ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc
from woke.ir.utils import IrInitTuple


class IndexRangeAccess(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcIndexRangeAccess
    _parent: SolidityAbc  # TODO: make this more specific

    _base_expression: ExpressionAbc
    _start_expression: Optional[ExpressionAbc]
    _end_expression: Optional[ExpressionAbc]

    def __init__(
        self,
        init: IrInitTuple,
        index_range_access: SolcIndexRangeAccess,
        parent: SolidityAbc,
    ):
        super().__init__(init, index_range_access, parent)
        self._base_expression = ExpressionAbc.from_ast(
            init, index_range_access.base_expression, self
        )

        if index_range_access.start_expression is None:
            self._start_expression = None
        else:
            self._start_expression = ExpressionAbc.from_ast(
                init, index_range_access.start_expression, self
            )

        if index_range_access.end_expression is None:
            self._end_expression = None
        else:
            self._end_expression = ExpressionAbc.from_ast(
                init, index_range_access.end_expression, self
            )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._base_expression
        if self._start_expression is not None:
            yield from self._start_expression
        if self._end_expression is not None:
            yield from self._end_expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def base_expression(self) -> ExpressionAbc:
        return self._base_expression

    @property
    def start_expression(self) -> Optional[ExpressionAbc]:
        return self._start_expression

    @property
    def end_expression(self) -> Optional[ExpressionAbc]:
        return self._end_expression

    @property
    def is_ref_to_state_variable(self) -> bool:
        # index range access in only supported for dynamic calldata arrays
        return False

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = self.base_expression.modifies_state
        if self.start_expression is not None:
            ret |= self.start_expression.modifies_state
        if self.end_expression is not None:
            ret |= self.end_expression.modifies_state
        return ret



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from woke.ir.ast import SolcYulFunctionCall, SolcYulIdentifier, SolcYulLiteral
from woke.ir.utils import IrInitTuple

from .abc import YulAbc
from .identifier import YulIdentifier
from .literal import YulLiteral

if TYPE_CHECKING:
    from .assignment import YulAssignment
    from .expression_statement import YulExpressionStatement
    from .for_loop import YulForLoop
    from .if_statement import YulIf
    from .switch import YulSwitch
    from .variable_declaration import YulVariableDeclaration


class YulFunctionCall(YulAbc):
    """
    TBD
    """

    _parent: Union[
        YulAssignment,
        YulExpressionStatement,
        YulForLoop,
        YulIf,
        YulSwitch,
        YulVariableDeclaration,
        YulFunctionCall,
    ]
    _arguments: List[Union["YulFunctionCall", YulIdentifier, YulLiteral]]
    _function_name: YulIdentifier

    def __init__(
        self, init: IrInitTuple, function_call: SolcYulFunctionCall, parent: YulAbc
    ):
        super().__init__(init, function_call, parent)
        self._function_name = YulIdentifier(init, function_call.function_name, self)
        self._arguments = []
        for argument in function_call.arguments:
            if isinstance(argument, SolcYulFunctionCall):
                self._arguments.append(YulFunctionCall(init, argument, self))
            elif isinstance(argument, SolcYulIdentifier):
                self._arguments.append(YulIdentifier(init, argument, self))
            elif isinstance(argument, SolcYulLiteral):
                self._arguments.append(YulLiteral(init, argument, self))
            else:
                assert False, f"Unexpected type: {type(argument)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._function_name
        for argument in self._arguments:
            yield from argument

    @property
    def parent(
        self,
    ) -> Union[
        YulAssignment,
        YulExpressionStatement,
        YulForLoop,
        YulIf,
        YulSwitch,
        YulVariableDeclaration,
        YulFunctionCall,
    ]:
        return self._parent

    @property
    def arguments(
        self,
    ) -> Tuple[Union["YulFunctionCall", YulIdentifier, YulLiteral], ...]:
        return tuple(self._arguments)

    @property
    def function_name(self) -> YulIdentifier:
        return self._function_name



from __future__ import annotations

from typing import List, Optional

import woke.ir as ir
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


def _check_variable_assigned_global_symbol(
    identifier: ir.Identifier, symbol: ir.enums.GlobalSymbolsEnum
):
    from woke.analysis.utils import (
        expression_is_global_symbol,
        get_function_definition_from_expression,
    )

    ident_statement = identifier
    while ident_statement.parent and not isinstance(ident_statement, ir.StatementAbc):
        ident_statement = ident_statement.parent
    if ident_statement is None:
        return False

    fn_dec = get_function_definition_from_expression(identifier)
    if fn_dec is None or fn_dec.body is None:
        return False

    assigned_val = None
    for stmt in fn_dec.body.statements_iter():
        if stmt == ident_statement:
            break
        if isinstance(stmt, ir.VariableDeclarationStatement):
            for i, dec in enumerate(stmt.declarations):
                if dec is not None and dec.name == identifier.name:
                    if isinstance(stmt.initial_value, ir.TupleExpression) and len(
                        stmt.declarations
                    ) == len(stmt.initial_value.components):
                        assigned_val = stmt.initial_value.components[i]
                    else:
                        assigned_val = stmt.initial_value
        elif (
            isinstance(stmt, ir.ExpressionStatement)
            and isinstance(stmt.expression, ir.Assignment)
            and isinstance(stmt.expression.left_expression, ir.Identifier)
        ):
            if (
                stmt.expression.left_expression.referenced_declaration
                == identifier.referenced_declaration
            ):
                assigned_val = stmt.expression.right_expression

    if not assigned_val:
        return False

    if expression_is_global_symbol(assigned_val, symbol):
        return True
    return False


def _check_unsafe_usage(node: ir.MemberAccess) -> Optional[DetectorResult]:
    from woke.analysis.utils import expression_is_global_symbol

    if node.parent and node.parent.parent:
        np = node.parent
        npp = node.parent.parent
        if (
            isinstance(np, ir.BinaryOperation)
            and np.operator == ir.enums.BinaryOpOperator.EQ
        ):
            if isinstance(npp, ir.FunctionCall) or isinstance(npp, ir.IfStatement):
                tx_origin_ma = (
                    np.right_expression
                    if np.right_expression == node
                    else np.left_expression
                )
                other = (
                    np.left_expression
                    if np.left_expression != tx_origin_ma
                    else np.right_expression
                )
                if isinstance(
                    tx_origin_ma, ir.MemberAccess
                ) and expression_is_global_symbol(
                    tx_origin_ma, ir.enums.GlobalSymbolsEnum.TX_ORIGIN
                ):
                    if isinstance(
                        other, ir.MemberAccess
                    ) and expression_is_global_symbol(
                        other, ir.enums.GlobalSymbolsEnum.MSG_SENDER
                    ):
                        return None
                    if (
                        isinstance(other, ir.Identifier)
                        and not other.is_ref_to_state_variable
                        and _check_variable_assigned_global_symbol(
                            other, ir.enums.GlobalSymbolsEnum.MSG_SENDER
                        )
                    ):
                        return None

        if isinstance(np, ir.IndexAccess):
            if (
                isinstance(npp, ir.BinaryOperation)
                and (npp.left_expression == np or npp.right_expression == np)
                and npp.operator == ir.enums.BinaryOpOperator.LT
            ):
                other_expr = (
                    npp.right_expression
                    if npp.left_expression == np
                    else npp.left_expression
                )
                if isinstance(
                    other_expr, ir.MemberAccess
                ) and expression_is_global_symbol(
                    other_expr, ir.enums.GlobalSymbolsEnum.BLOCK_NUMBER
                ):
                    return None
            elif (
                isinstance(npp, ir.Assignment)
                and isinstance(npp.right_expression, ir.MemberAccess)
                and expression_is_global_symbol(
                    npp.right_expression, ir.enums.GlobalSymbolsEnum.BLOCK_NUMBER
                )
            ):
                return None

    return DetectorResult(
        Detection(node, "tx.origin used in an unsafe manner"),
        confidence=DetectionConfidence.LOW,
        impact=DetectionImpact.MEDIUM,
    )


def detect_unsafe_tx_origin(node: ir.MemberAccess) -> Optional[DetectorResult]:
    from woke.analysis.utils import expression_is_global_symbol

    if expression_is_global_symbol(node, ir.enums.GlobalSymbolsEnum.TX_ORIGIN):
        return _check_unsafe_usage(node)


class UnsafeTxOriginDetector(Detector):
    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_member_access(self, node: ir.MemberAccess):
        res = detect_unsafe_tx_origin(node)
        if res:
            self._detections.add(res)

    @detector.command("unsafe-tx-origin")
    def cli(self):
        """
        Detects unsafe usage of tx.origin.

        Every usage of tx.origin us unsafe expect for the following:
        - tx.origin == msg.sender and it's varieties
        - timestamp[tx.origin] < block.number
        """
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from woke.ir.ast import SolcYulLiteral
from woke.ir.enums import YulLiteralValueKind
from woke.ir.utils import IrInitTuple

from .abc import YulAbc

if TYPE_CHECKING:
    from .assignment import YulAssignment
    from .case_statement import YulCase
    from .expression_statement import YulExpressionStatement
    from .for_loop import YulForLoop
    from .function_call import YulFunctionCall
    from .if_statement import YulIf
    from .switch import YulSwitch
    from .variable_declaration import YulVariableDeclaration


class YulLiteral(YulAbc):
    """
    TBD
    """

    _parent: Union[
        YulAssignment,
        YulExpressionStatement,
        YulForLoop,
        YulIf,
        YulSwitch,
        YulVariableDeclaration,
        YulFunctionCall,
        YulCase,
    ]
    _kind: YulLiteralValueKind
    _type: str
    _value: Optional[str]
    _hex_value: Optional[str]

    def __init__(self, init: IrInitTuple, literal: SolcYulLiteral, parent: YulAbc):
        super().__init__(init, literal, parent)
        self._kind = literal.kind
        self._type = literal.type
        self._value = literal.value
        self._hex_value = literal.hex_value

    @property
    def parent(
        self,
    ) -> Union[
        YulAssignment,
        YulExpressionStatement,
        YulForLoop,
        YulIf,
        YulSwitch,
        YulVariableDeclaration,
        YulFunctionCall,
        YulCase,
    ]:
        return self._parent

    @property
    def kind(self) -> YulLiteralValueKind:
        return self._kind

    @property
    def type(self) -> str:
        return self._type

    @property
    def value(self) -> Optional[str]:
        return self._value

    @property
    def hex_value(self) -> Optional[str]:
        return self._hex_value



from __future__ import annotations

from collections import deque
from pathlib import Path
from typing import TYPE_CHECKING, Deque, Iterator, Optional, Set, Tuple, Union

from intervaltree import IntervalTree

from woke.ir.types import Contract, Enum, Struct, UserDefinedValueType

from ..reference_resolver import CallbackParams

if TYPE_CHECKING:
    from ..declarations.variable_declaration import VariableDeclaration
    from ..expressions.new_expression import NewExpression
    from ..meta.inheritance_specifier import InheritanceSpecifier
    from ..meta.override_specifier import OverrideSpecifier
    from ..meta.using_for_directive import UsingForDirective
    from ..meta.source_unit import SourceUnit
    from ..type_names.array_type_name import ArrayTypeName
    from ..type_names.mapping import Mapping

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import AstNodeId, SolcUserDefinedTypeName
from woke.ir.declarations.abc import DeclarationAbc
from woke.ir.meta.identifier_path import (
    IDENTIFIER_RE,
    IdentifierPath,
    IdentifierPathPart,
)
from woke.ir.type_names.abc import TypeNameAbc
from woke.ir.utils import IrInitTuple


class UserDefinedTypeName(TypeNameAbc):
    """
    User defined type name represents a name path to a user defined type. Path parts are separated by dots.
    In Solidity 0.8.0 a new IR node ([IdentifierPath][woke.ir.meta.identifier_path.IdentifierPath]) was introduced to replace [UserDefinedTypeName][woke.ir.type_names.user_defined_type_name.UserDefinedTypeName] in some cases.

    !!! example
        A user defined type name can be used:

        - inside a [VariableDeclaration][woke.ir.declarations.variable_declaration.VariableDeclaration]:
            - `:::solidity Interface.Struct` in line 18,
            - `:::solidity Interface.Enum` in line 26,
        - inside a [NewExpression][woke.ir.expressions.new_expression.NewExpression]:
            - `:::solidity Contract` in line 20,
        - inside an [InheritanceSpecifier][woke.ir.meta.inheritance_specifier.InheritanceSpecifier]:
            - `:::solidity Interface` in line 23,
        - inside an [OverrideSpecifier][woke.ir.meta.override_specifier.OverrideSpecifier]:
            - `:::solidity Interface` in line 30,
        - inside a [UsingForDirective][woke.ir.meta.using_for_directive.UsingForDirective]:
            - `:::solidity Lib` in line 24,
            - `:::solidity Interface.Struct` in line 24,
        - inside an [ArrayTypeName][woke.ir.type_names.array_type_name.ArrayTypeName]:
            - `:::solidity Interface.Enum` in line 27,
        - inside a [Mapping][woke.ir.type_names.mapping.Mapping]:
            - both occurrences of `:::solidity Interface.Enum` in line 28.

        ```solidity linenums="1"
        pragma solidity 0.7;

        interface Interface {
            enum Enum {
                READY,
                WAITING
            }

            struct Struct {
                uint a;
            }

            function foo() external;
        }

        library Lib {}

        function tmp(Interface.Struct memory s) {
            s.a = 5;
            new Contract();
        }

        contract Contract is Interface {
            using Lib for Interface.Struct;

            Interface.Enum state;
            Interface.Enum[] states;
            mapping(Interface.Enum => Interface.Enum) map;

            function foo() external override(Interface) {
            }
        }
        ```
    """

    _ast_node: SolcUserDefinedTypeName
    _parent: Union[
        VariableDeclaration,
        NewExpression,
        InheritanceSpecifier,
        OverrideSpecifier,
        UsingForDirective,
        ArrayTypeName,
        Mapping,
    ]

    _referenced_declaration_id: AstNodeId
    _contract_scope_id: Optional[AstNodeId]
    _name: Optional[str]
    _path_node: Optional[IdentifierPath]
    _parts: Optional[IntervalTree]

    def __init__(
        self,
        init: IrInitTuple,
        user_defined_type_name: SolcUserDefinedTypeName,
        parent: SolidityAbc,
    ):
        super().__init__(init, user_defined_type_name, parent)
        self._name = user_defined_type_name.name
        self._referenced_declaration_id = user_defined_type_name.referenced_declaration
        assert self._referenced_declaration_id >= 0
        self._contract_scope_id = user_defined_type_name.contract_scope

        if user_defined_type_name.path_node is None:
            self._path_node = None
            self._reference_resolver.register_post_process_callback(self._post_process)
        else:
            self._path_node = IdentifierPath(
                init, user_defined_type_name.path_node, self
            )
            self._parts = None

    def _post_process(self, callback_params: CallbackParams):
        def find_referenced_source_unit(
            searched_name: str, start_source_unit: SourceUnit
        ) -> SourceUnit:
            source_units_queue: Deque[SourceUnit] = deque([start_source_unit])
            processed_source_units: Set[Path] = {start_source_unit.file}
            referenced_declaration = None

            while source_units_queue and referenced_declaration is None:
                source_unit = source_units_queue.popleft()

                for import_ in source_unit.imports:
                    if import_.unit_alias == searched_name:
                        referenced_declaration = callback_params.source_units[
                            import_.imported_file
                        ]
                        break
                    for symbol_alias in import_.symbol_aliases:
                        if symbol_alias.local == searched_name:
                            ref = symbol_alias.foreign.referenced_declaration
                            assert isinstance(ref, SourceUnit)
                            referenced_declaration = ref

                    if referenced_declaration is not None:
                        break

                    if import_.imported_file not in processed_source_units:
                        source_units_queue.append(
                            callback_params.source_units[import_.imported_file]
                        )
                        processed_source_units.add(import_.imported_file)

            assert referenced_declaration is not None
            return referenced_declaration

        from ..meta.source_unit import SourceUnit

        matches = list(IDENTIFIER_RE.finditer(self._source))
        groups_count = len(matches)
        assert groups_count > 0

        self._parts = IntervalTree()
        start_source_unit = callback_params.source_units[self._file]

        ref = self.referenced_declaration
        refs = []
        for _ in range(groups_count):
            refs.append(ref)
            if ref is not None:
                ref = ref.parent

        for match, ref in zip(matches, reversed(refs)):
            name = match.group(0).decode("utf-8")

            if ref is None:
                start_source_unit = find_referenced_source_unit(name, start_source_unit)
                referenced_node = start_source_unit
            elif isinstance(ref, (DeclarationAbc, SourceUnit)):
                referenced_node = ref
            else:
                raise TypeError(
                    f"Unexpected type of referenced declaration: {type(ref)}"
                )

            node_path_order = self._reference_resolver.get_node_path_order(
                AstNodeId(referenced_node.ast_node_id),
                referenced_node.cu_hash,
            )
            referenced_node_id = (
                self._reference_resolver.get_ast_id_from_cu_node_path_order(
                    node_path_order, self._cu_hash
                )
            )

            start = self.byte_location[0] + match.start()
            end = self.byte_location[0] + match.end()
            self._parts[start:end] = IdentifierPathPart(
                self,
                (start, end),
                name,
                referenced_node_id,
                self._reference_resolver,
                self._cu_hash,
                self._file,
            )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        if self._path_node is not None:
            yield from self._path_node

    @property
    def parent(
        self,
    ) -> Union[
        VariableDeclaration,
        NewExpression,
        InheritanceSpecifier,
        OverrideSpecifier,
        UsingForDirective,
        ArrayTypeName,
        Mapping,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def type(self) -> Union[Contract, Struct, Enum, UserDefinedValueType]:
        """
        Returns:
            Type description.
        """
        t = super().type
        assert isinstance(t, (Contract, Struct, Enum, UserDefinedValueType))
        return t

    @property
    def name(self) -> str:
        """
        !!! note
            Should be the same as [source][woke.ir.abc.IrAbc.source] and is the same as [path_node.name][woke.ir.meta.identifier_path.IdentifierPath.name] if [path_node][woke.ir.type_names.user_defined_type_name.UserDefinedTypeName.path_node] is not `None`.
        Returns:
            Name of the user defined type as it appears in the source code.
        """
        if self._name is None:
            assert self._path_node is not None
            self._name = self._path_node.name
        return self._name

    @property
    def identifier_path_parts(self) -> Tuple[IdentifierPathPart, ...]:
        """
        Returns:
            Parts of the user defined type name.
        """
        if self._path_node is not None:
            return self._path_node.identifier_path_parts

        assert self._parts is not None
        return tuple(interval.data for interval in sorted(self._parts))

    def identifier_path_part_at(self, byte_offset: int) -> Optional[IdentifierPathPart]:
        """
        Args:
            byte_offset: Byte offset in the source file.
        Returns:
            Identifier path part at the given byte offset, if any.
        """
        if self._path_node is not None:
            return self._path_node.identifier_path_part_at(byte_offset)

        assert self._parts is not None
        intervals = self._parts.at(byte_offset)
        assert len(intervals) <= 1
        if len(intervals) == 0:
            return None
        return intervals.pop().data

    @property
    def referenced_declaration(self) -> DeclarationAbc:
        """
        Returns:
            Declaration IR node referenced by this user defined type name.
        """
        node = self._reference_resolver.resolve_node(
            self._referenced_declaration_id, self._cu_hash
        )
        assert isinstance(node, DeclarationAbc)
        return node

    @property
    def path_node(self) -> Optional[IdentifierPath]:
        """
        Always present since Solidity 0.8.0. If not `None`, it represents the same source code as this node ([byte_location][woke.ir.abc.IrAbc.byte_location] properties are the same) and references the same declaration.
        Returns:
            Identifier path IR node.
        """
        return self._path_node



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

from woke.ir.ast import (
    SolcYulFunctionCall,
    SolcYulIdentifier,
    SolcYulLiteral,
    SolcYulVariableDeclaration,
)

from ..utils import IrInitTuple
from .abc import YulAbc, YulStatementAbc
from .function_call import YulFunctionCall
from .identifier import YulIdentifier
from .literal import YulLiteral
from .typed_name import YulTypedName

if TYPE_CHECKING:
    from .block import YulBlock


class YulVariableDeclaration(YulStatementAbc):
    """
    TBD
    """

    _parent: YulBlock
    _variables: List[YulTypedName]
    _value: Optional[Union[YulFunctionCall, YulIdentifier, YulLiteral]]

    def __init__(
        self,
        init: IrInitTuple,
        variable_declaration: SolcYulVariableDeclaration,
        parent: YulAbc,
    ):
        super().__init__(init, variable_declaration, parent)
        self._variables = [
            YulTypedName(init, variable, self)
            for variable in variable_declaration.variables
        ]
        if variable_declaration.value is None:
            self._value = None
        elif isinstance(variable_declaration.value, SolcYulFunctionCall):
            self._value = YulFunctionCall(init, variable_declaration.value, self)
        elif isinstance(variable_declaration.value, SolcYulIdentifier):
            self._value = YulIdentifier(init, variable_declaration.value, self)
        elif isinstance(variable_declaration.value, SolcYulLiteral):
            self._value = YulLiteral(init, variable_declaration.value, self)
        else:
            assert False, f"Unexpected type: {type(variable_declaration.value)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        for variable in self._variables:
            yield from variable
        if self._value is not None:
            yield from self._value

    @property
    def parent(self) -> YulBlock:
        return self._parent

    @property
    def variables(self) -> Tuple[YulTypedName, ...]:
        return tuple(self._variables)

    @property
    def value(self) -> Optional[Union[YulFunctionCall, YulIdentifier, YulLiteral]]:
        return self._value



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union

import woke.ir as ir
import woke.ir.types as types
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


@lru_cache(maxsize=2048)
def _detect_slot_value(expr: ir.ExpressionAbc) -> Optional[Detection]:

    val = expr
    while True:
        if isinstance(val, ir.Literal) and val.kind in (
            ir.enums.LiteralKind.NUMBER,
            ir.enums.LiteralKind.STRING,
        ):
            return Detection(val, "Detected slot in upgrade contract")
        elif isinstance(val, ir.BinaryOperation) and isinstance(
            val.left_expression, ir.FunctionCall
        ):
            val = val.left_expression
        elif isinstance(val, ir.BinaryOperation) and isinstance(
            val.right_expression, ir.FunctionCall
        ):
            val = val.right_expression
        elif (
            isinstance(val, ir.FunctionCall)
            and val.function_called != ir.enums.GlobalSymbolsEnum.KECCAK256
            and len(val.arguments) == 1
        ):
            val = val.arguments[0]
        else:
            break
    if (
        isinstance(val, ir.FunctionCall)
        and val.function_called == ir.enums.GlobalSymbolsEnum.KECCAK256
    ):
        return Detection(expr, "Detected slot through keccak256 in upgrade contract")
    return None


@lru_cache(maxsize=2048)
def _detect_implementation_fn(fn: ir.FunctionDefinition) -> List[Detection]:
    from woke.analysis.utils import get_function_implementations

    dets = []
    if fn.body is None:
        for fn_impl in get_function_implementations(fn):
            if not isinstance(fn_impl, ir.FunctionDefinition):
                continue
            dets.extend(_detect_implementation_fn(fn_impl))
        return dets

    addr_detected = False
    for param in fn.return_parameters.parameters:
        if isinstance(param.type, types.Address):
            addr_detected = True
    if not addr_detected:
        return dets

    dets = []
    for node in fn.body:
        if isinstance(node, ir.Return) and node.expression is not None:
            for det in _detect_slot_variable(node.expression):
                dets.append(
                    Detection(
                        fn,
                        "Detected slot usage in implementation function",
                        subdetections=(det,),
                        lsp_range=fn.name_location,
                    )
                )
        elif isinstance(node, ir.FunctionCall) and isinstance(
            node.function_called, ir.FunctionDefinition
        ):
            for det in _detect_implementation_fn(node.function_called):
                dets.append(
                    Detection(
                        fn,
                        "Detected implementation function",
                        subdetections=(det,),
                        lsp_range=fn.name_location,
                    )
                )
        elif isinstance(node, ir.InlineAssembly) and node.yul_block is not None:
            for det in _check_assembly_uses_slot_variable(node.yul_block):
                dets.append(
                    Detection(
                        fn,
                        "Detected slot usage in implementation function",
                        subdetections=(det,),
                        lsp_range=fn.name_location,
                    )
                )
    return dets


def _detect_slot_variable(
    expr: Union[ir.ExpressionAbc, ir.DeclarationAbc, ir.StatementAbc],
    value: Optional[ir.ExpressionAbc] = None,
    visited=None,
) -> List[Detection]:
    dets = []
    if visited is None:
        visited = set()
    if (expr, value) in visited:
        return dets
    visited.add((expr, value))

    if isinstance(expr, ir.TupleExpression):
        for t in expr:
            if isinstance(t, ir.ExpressionAbc) or isinstance(t, ir.DeclarationAbc):
                dets.extend(_detect_slot_variable(t, visited=visited))
    elif isinstance(expr, ir.VariableDeclaration):
        val = expr.value if expr.value is not None else value
        if (
            expr.is_state_variable
            and isinstance(expr.type, types.FixedBytes)
            and expr.type.bytes_count == 32
        ):
            if val is not None:
                det = _detect_slot_value(val)
                if det is not None:
                    dets.append(det)
        elif isinstance(expr.type, types.Address):
            if val is not None:
                det = _detect_slot_value(val)
                if det is not None:
                    dets.append(det)
                if isinstance(val, ir.FunctionCall) and isinstance(
                    val.function_called, ir.FunctionDefinition
                ):
                    for arg in val.arguments:
                        dets.extend(_detect_slot_variable(arg, visited=visited))
                    dets.extend(_detect_implementation_fn(val.function_called))
            elif expr.parent is not None and isinstance(
                expr.parent, ir.VariableDeclarationStatement
            ):
                dets.extend(_detect_slot_variable(expr.parent, visited=visited))
    elif (
        isinstance(expr, ir.VariableDeclarationStatement)
        and expr.initial_value is not None
        and (
            (
                isinstance(expr.initial_value, ir.TupleExpression)
                and len(expr.declarations) == len(expr.initial_value.components)
            )
            or (len(expr.declarations) == 1)
        )
    ):
        if isinstance(expr.initial_value, ir.TupleExpression):
            vals = expr.initial_value.components
        elif isinstance(expr.initial_value, ir.ExpressionAbc):
            vals = [expr.initial_value]
        elif isinstance(expr.initial_value, List):
            vals = expr.initial_value
        else:
            return dets
        for i in range(len(vals)):
            ini_val = vals[i]
            if isinstance(ini_val, ir.Identifier) and isinstance(
                ini_val.referenced_declaration, ir.VariableDeclaration
            ):
                dets.extend(
                    _detect_slot_variable(
                        ini_val.referenced_declaration, visited=visited
                    )
                )
            elif isinstance(ini_val, ir.ExpressionAbc) and ini_val is not None:
                decl = expr.declarations[i]
                if decl is not None:
                    dets.extend(_detect_slot_variable(decl, ini_val, visited=visited))
    elif (
        isinstance(expr, ir.Identifier)
        and isinstance(expr.referenced_declaration, ir.VariableDeclaration)
        and expr.is_ref_to_state_variable
    ):
        dets.extend(_detect_slot_variable(expr.referenced_declaration, visited=visited))
    elif isinstance(expr, ir.MemberAccess):
        for n in expr:
            if n == expr:
                continue
            if isinstance(n, ir.ExpressionAbc) or isinstance(n, ir.DeclarationAbc):
                dets.extend(_detect_slot_variable(n, visited=visited))
    return dets


def _detect_delegatecall_to_slot_variable(
    fn: ir.FunctionDefinition,
    callargs: Optional[Tuple[Tuple[ir.VariableDeclaration, ir.ExpressionAbc], ...]],
    visited=None,
) -> List[Detection]:
    from woke.analysis.utils import (
        get_function_implementations,
        pair_function_call_arguments,
    )

    dets = []
    if visited is None:
        visited = []
    if fn in visited:
        return dets
    visited.append(fn)

    if fn.body is None:
        for impl in get_function_implementations(fn):
            if isinstance(impl, ir.FunctionDefinition):
                dets.extend(
                    _detect_delegatecall_to_slot_variable(impl, callargs, visited)
                )
        return dets

    for node in fn.body:
        if isinstance(node, ir.FunctionCall):
            if isinstance(node.function_called, ir.FunctionDefinition):
                for det in _detect_delegatecall_to_slot_variable(
                    node.function_called,
                    pair_function_call_arguments(node.function_called, node),
                    visited,
                ):
                    dets.append(Detection(node, "Detected call", subdetections=(det,)))
            elif (
                node.function_called == ir.enums.GlobalSymbolsEnum.ADDRESS_DELEGATECALL
                and len(node.arguments) == 1
            ):
                if isinstance(node.expression, ir.MemberAccess) and isinstance(
                    node.expression.referenced_declaration, ir.DeclarationAbc
                ):
                    for det in _detect_slot_variable(
                        node.expression.referenced_declaration
                    ):
                        dets.append(
                            Detection(
                                fn,
                                "Detected delegate call to an implementation slot",
                                subdetections=(
                                    Detection(
                                        node, "Detected call", subdetections=(det,)
                                    ),
                                ),
                                lsp_range=fn.name_location,
                            )
                        )
        elif (
            isinstance(node, ir.YulFunctionCall)
            and node.function_name.name == "delegatecall"
            and len(node.arguments) > 1
        ):
            arg = node.arguments[1]  # type: ignore
            if isinstance(arg, ir.YulIdentifier) and arg.external_reference is not None:
                ref_decl = arg.external_reference.referenced_declaration
                assembly_dets = []
                if (
                    ref_decl in fn.parameters.parameters
                    and len(fn.parameters.parameters) > 0
                    and callargs is not None
                    and fn.parameters.parameters.index(ref_decl) < len(callargs)
                ):
                    assembly_dets = _detect_slot_variable(
                        ref_decl, callargs[fn.parameters.parameters.index(ref_decl)][1]
                    )
                else:
                    assembly_dets = _detect_slot_variable(ref_decl)
                for det in assembly_dets:
                    dets.append(
                        Detection(
                            fn,
                            "Detected assembly delegate call using an implementation slot",
                            subdetections=(
                                Detection(
                                    node,
                                    "Detected assembly call",
                                    subdetections=(
                                        Detection(
                                            ref_decl,
                                            "Detected reference with implementation value passed in callargs",
                                            subdetections=(det,),
                                        ),
                                    ),
                                ),
                            ),
                            lsp_range=fn.name_location,
                        )
                    )
    return dets


def _check_assembly_uses_slot_variable(block: ir.YulBlock) -> List[Detection]:
    dets = []
    if block.statements is None:
        return dets

    for yul in block:
        if isinstance(yul, ir.YulIdentifier):
            if yul.external_reference is not None:
                for det in _detect_slot_variable(
                    yul.external_reference.referenced_declaration
                ):
                    dets.append(
                        Detection(
                            yul,
                            "Detected slot variable usage in assembly",
                            subdetections=(det,),
                        )
                    )
    return dets


def detect_fallback(fn: ir.FunctionDefinition) -> List[Detection]:
    dets = []
    if fn.kind != ir.enums.FunctionKind.FALLBACK:
        return dets

    if fn.body is None:
        return dets

    for det in _detect_delegatecall_to_slot_variable(fn, None):
        dets.append(
            Detection(
                fn,
                "Detected fallback function with delegate call to an implementation slot",
                subdetections=(det,),
                lsp_range=fn.name_location,
            )
        )
    return dets


def get_last_detection_node(det: Detection) -> ir.IrAbc:
    if len(det.subdetections) == 0:
        return det.ir_node
    return get_last_detection_node(det.subdetections[0])


def detect_slot_usage(fn: ir.FunctionDefinition, visited=None) -> List[Detection]:
    from woke.analysis.utils import get_function_implementations

    dets = []
    if visited is None:
        visited = []
    if fn in visited:
        return dets
    visited.append(fn)
    if fn.body is None:
        return dets

    if not fn.implemented:
        for impl_fn in get_function_implementations(fn):
            if isinstance(impl_fn, ir.FunctionDefinition):
                dets.extend(detect_slot_usage(impl_fn, visited=visited))
        return dets
    for node in fn.body:
        if (
            isinstance(node, ir.Identifier) or isinstance(node, ir.MemberAccess)
        ) and isinstance(node.referenced_declaration, ir.DeclarationAbc):
            for det in _detect_slot_variable(node.referenced_declaration):
                dets.append(
                    Detection(
                        node,
                        "Detected slot variable usage",
                        subdetections=(det,),
                    )
                )
        elif isinstance(node, ir.FunctionCall) and isinstance(
            node.function_called, ir.FunctionDefinition
        ):
            for det in detect_slot_usage(node.function_called, visited=visited):
                dets.append(
                    Detection(
                        node,
                        "Detected slot variable usage",
                        subdetections=(det,),
                    )
                )
        elif isinstance(node, ir.InlineAssembly) and node.yul_block is not None:
            for det in _check_assembly_uses_slot_variable(node.yul_block):
                dets.append(
                    Detection(
                        node,
                        "Detected slot variable usage",
                        subdetections=(det,),
                    )
                )
    return dets


def detect_selector_clashes(
    proxy_contract: ir.ContractDefinition,
    impl_contract: ir.ContractDefinition,
    proxy_detection: Detection,
    impl_detection: Detection,
) -> List[DetectorResult]:
    fn_whitelist = ["implementation"]

    proxy_selectors = {}
    for c in proxy_contract.linearized_base_contracts:
        for f in c.functions + c.declared_variables:
            if f.function_selector is not None:
                proxy_selectors[f.function_selector] = f

    impl_selectors = {}
    for c in impl_contract.linearized_base_contracts:
        for f in c.functions + c.declared_variables:
            if f.function_selector is not None:
                impl_selectors[f.function_selector] = f

    clashes = []
    for proxy_sel, proxy_fn in proxy_selectors.items():
        if (
            isinstance(proxy_fn, ir.FunctionDefinition)
            and proxy_fn.name in fn_whitelist
        ):
            continue
        if proxy_sel in impl_selectors:
            clashes.append(
                DetectorResult(
                    Detection(
                        proxy_fn,
                        "Detected selector clash with implementation contract",
                        subdetections=(
                            Detection(
                                impl_selectors[proxy_sel],
                                "Implementation function with same selector",
                                subdetections=(impl_detection,),
                                lsp_range=impl_selectors[proxy_sel].name_location
                                if isinstance(
                                    impl_selectors[proxy_sel], ir.FunctionDefinition
                                )
                                else None,
                            ),
                            proxy_detection,
                        ),
                        lsp_range=proxy_fn.name_location,
                    ),
                    confidence=DetectionConfidence.MEDIUM,
                    impact=DetectionImpact.MEDIUM,
                )
            )
    return clashes


class ProxyContractSelectorClashDetector(Detector):
    _proxy_detections: Set[Tuple[ir.ContractDefinition, Detection]]
    _proxy_associated_contracts: Set[ir.ContractDefinition]
    _implementation_slots_detections: Set[Tuple[ir.ContractDefinition, Detection]]
    _implementation_slots: Dict[ir.VariableDeclaration, List[ir.ContractDefinition]]

    def __init__(self):
        self._proxy_detections: Set[Tuple[ir.ContractDefinition, Detection]] = set()
        self._proxy_associated_contracts: Set[ir.ContractDefinition] = set()
        self._implementation_slots_detections: Set[
            Tuple[ir.ContractDefinition, Detection]
        ] = set()
        self._implementation_slots: Dict[
            ir.VariableDeclaration, List[ir.ContractDefinition]
        ] = {}

    def detect(self) -> List[DetectorResult]:
        detections = []

        base_proxy_contracts = set()
        for (contract, _) in self._proxy_detections:
            for c in contract.linearized_base_contracts:
                if c == contract:
                    continue
                base_proxy_contracts.add(c)

        proxy_detections = {}
        for (contract, det) in self._proxy_detections:
            if contract not in base_proxy_contracts:
                proxy_detections[contract] = det

        base_impl_contracts = set()
        for (contract, _) in self._implementation_slots_detections:
            for c in contract.linearized_base_contracts:
                if c == contract:
                    continue
                base_impl_contracts.add(c)

        checked_pairs = set()
        impl_detections_contracts = set()
        for (contract, det) in self._implementation_slots_detections:
            impl_slot = get_last_detection_node(det).parent
            if (
                contract not in self._proxy_associated_contracts
                and contract not in base_impl_contracts
                and impl_slot in self._implementation_slots
                and contract not in impl_detections_contracts
            ):
                for proxy_contract in self._implementation_slots[impl_slot]:
                    if (
                        proxy_contract,
                        contract,
                    ) in checked_pairs or proxy_contract not in proxy_detections:
                        continue
                    checked_pairs.add((proxy_contract, contract))
                    impl_det = Detection(
                        contract,
                        f"Detected implementation contract with slot used in proxy contract {proxy_contract.name}",
                        subdetections=(det,),
                        lsp_range=contract.name_location,
                    )

                    dets = detect_selector_clashes(
                        proxy_contract,
                        contract,
                        proxy_detections[proxy_contract],
                        impl_det,
                    )
                    if len(dets) > 0:
                        detections.extend(dets)
        return list(detections)

    def visit_contract_definition(self, node: ir.ContractDefinition):
        searched_fns = set()
        if node not in self._proxy_associated_contracts:
            for b in node.linearized_base_contracts:
                for f in b.functions:
                    if f.implemented and f not in searched_fns:
                        searched_fns.add(f)
                        dets = detect_fallback(f)
                        if len(dets) > 0:
                            self._proxy_detections.add(
                                (
                                    node,
                                    Detection(
                                        node,
                                        "Detected proxy contract",
                                        subdetections=tuple(dets),
                                        lsp_range=node.name_location,
                                    ),
                                )
                            )
                            for det in dets:
                                last_det_node = get_last_detection_node(det)
                                if (
                                    isinstance(
                                        last_det_node.parent, ir.VariableDeclaration
                                    )
                                    and last_det_node.parent is not None
                                ):
                                    if (
                                        last_det_node.parent
                                        not in self._implementation_slots
                                    ):
                                        self._implementation_slots[
                                            last_det_node.parent
                                        ] = []
                                    self._implementation_slots[
                                        last_det_node.parent
                                    ].append(node)
                                self._proxy_associated_contracts.add(node)
                                for bc in node.linearized_base_contracts:
                                    self._proxy_associated_contracts.add(bc)

        if node not in self._proxy_associated_contracts:
            for fn in node.functions:
                for det in detect_slot_usage(fn):
                    self._implementation_slots_detections.add((node, det))

    @detector.command(name="proxy-contract-selector-clashes")
    def cli(self):
        """
        Detects selector clashes in proxy and implementation contracts.
        Proxy contracts are detected based on fallback function and usage of slot variables and
        implementation contracts that use same slots as proxy contracts
        """
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Optional, Tuple

from woke.core.solidity_version import SolidityVersionRanges
from woke.ir.ast import SolcNode, SolidityNode
from woke.ir.reference_resolver import ReferenceResolver
from woke.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from woke.ir.meta.source_unit import SourceUnit


class IrAbc(ABC):
    """
    Base class for all IR nodes. Defines attributes and methods that are common to all Solidity and Yul IR nodes.

    IR model is built on top of the AST (Abstract Syntax Tree) output of the [solc compiler](https://docs.soliditylang.org/en/latest/using-the-compiler.html).

    Each IR node is associated with a [source code location][woke.ir.abc.IrAbc.byte_location] in a [source file][woke.ir.abc.IrAbc.file].
    This means that each IR node has a corresponding (typically non-empty) Solidity or Yul [source code][woke.ir.abc.IrAbc.source].

    !!! info
        Yul IR nodes can have empty source code. In the case of Solidity IR nodes, this should not happen.

    """

    _file: Path
    _source: bytes
    _ast_node: SolcNode
    _version_ranges: SolidityVersionRanges
    _parent: Optional[IrAbc]
    _depth: int
    _cu_hash: bytes
    _source_unit: SourceUnit
    _reference_resolver: ReferenceResolver

    def __init__(self, init: IrInitTuple, solc_node: SolcNode, parent: Optional[IrAbc]):
        self._file = init.file
        self._ast_node = solc_node
        self._version_ranges = init.cu.versions
        self._parent = parent
        if self._parent is not None:
            self._depth = self._parent.ast_tree_depth + 1
        else:
            self._depth = 0
        self._cu_hash = init.cu.hash

        assert init.source_unit is not None
        self._source_unit = init.source_unit
        self._reference_resolver = init.reference_resolver

        source_start = solc_node.src.byte_offset
        source_end = source_start + solc_node.src.byte_length
        self._source = init.source[source_start:source_end]
        if source_start != source_end:
            init.interval_tree[source_start:source_end] = self

    def __iter__(self) -> Iterator[IrAbc]:
        """
        Yields:
            Self and (recursively) all child IR nodes.
        """
        yield self

    @property
    @abstractmethod
    def parent(self) -> Optional[IrAbc]:
        """
        The parent node of this node. Can only be `None` for the root ([Source unit][woke.ir.meta.source_unit.SourceUnit]) node.

        Returns:
            Parent node of this node.
        """
        ...

    @property
    def file(self) -> Path:
        """
        The absolute path to the source file that contains this IR node. For a given IR node, all child and parent nodes have the same file path.
        Returns:
            Absolute path to the file containing this node.
        """
        return self._file

    @property
    @abstractmethod
    def ast_node(self) -> SolcNode:
        ...

    @property
    def cu_hash(self) -> bytes:
        return self._cu_hash

    @property
    def version_ranges(self) -> SolidityVersionRanges:
        """
        !!! example
            ```python
            if "0.8.0" in node.version_ranges:
                print("The given file can be compiled with solc 0.8.0")
            ```
        Returns:
            Object listing all `solc` versions that can be used to compile the file containing this node.
        """
        return self._version_ranges

    @property
    def ast_tree_depth(self) -> int:
        """
        The depth of this node in the AST tree. The root node ([Source unit][woke.ir.meta.source_unit.SourceUnit]) of each file has depth 0. Direct child nodes of a `node` have depth `{node}.ast_tree_depth + 1`.

        !!! tip
            Woke uses [interval trees](https://github.com/chaimleib/intervaltree) to get a list of all IR nodes at a given byte offset in a given file. This property can be used to sort these nodes by their depth in the AST tree and (for example) to choose the most nested one.

        Returns:
            Depth of this node in the AST tree, starting from 0.

        """
        return self._depth

    @property
    def byte_location(self) -> Tuple[int, int]:
        """
        The start and end byte offsets of this node in the source file. `{node}.byte_location[0]` is the start byte offset, `{node}.byte_location[1]` is the end byte offset.

        `{node}.byte_location[1]` is always greater than or equal to `{node}.byte_location[0]`.

        The byte location of a child node is typically a subrange of the byte location of its parent node.

        !!! info
            This is not true for [Structured documentation][woke.ir.meta.structured_documentation.StructuredDocumentation], where documentation strings must be located before a declaration.

        Returns:
            Tuple of the start and end byte offsets of this node in the source file.
        """
        return (
            self._ast_node.src.byte_offset,
            self._ast_node.src.byte_offset + self._ast_node.src.byte_length,
        )

    @property
    def source(self) -> str:
        """
        UTF-8 decoded source code from the [source file][woke.ir.abc.IrAbc.file] at the [byte offset][woke.ir.abc.IrAbc.byte_location] of this node.
        Returns:
            Solidity or Yul source code corresponding to this node.
        """
        return self._source.decode("utf-8")

    @property
    def source_unit(self) -> SourceUnit:
        """
        Returns:
            [source unit][woke.ir.meta.source_unit.SourceUnit] that contains this node.
        """
        return self._source_unit


class SolidityAbc(IrAbc, ABC):
    """
    Abstract base class for all Solidity IR nodes.
    """

    _ast_node: SolidityNode

    def __init__(
        self, init: IrInitTuple, solc_node: SolidityNode, parent: Optional[SolidityAbc]
    ):
        super().__init__(init, solc_node, parent)
        self._reference_resolver.register_node(self, solc_node.id, self._cu_hash)

    @property
    def ast_node(self) -> SolidityNode:
        return self._ast_node

    @property
    def ast_node_id(self) -> int:
        return self._ast_node.id



from collections import defaultdict


# https://stackoverflow.com/a/2912455
class KeyedDefaultDict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(
                key  # pyright: ignore reportGeneralTypeIssues
            )
            return ret



from __future__ import annotations

import re
from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

from ..meta.structured_documentation import StructuredDocumentation
from .abc import DeclarationAbc

if TYPE_CHECKING:
    from .contract_definition import ContractDefinition
    from ..meta.source_unit import SourceUnit

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcStructDefinition
from woke.ir.declarations.variable_declaration import VariableDeclaration
from woke.ir.enums import Visibility
from woke.ir.utils import IrInitTuple


class StructDefinition(DeclarationAbc):
    """
    Definition of a struct.

    !!! example
        ```solidity
        struct S {
            uint a;
            uint b;
        }
        ```
    """

    _ast_node: SolcStructDefinition
    _parent: Union[ContractDefinition, SourceUnit]

    _canonical_name: str
    _members: List[VariableDeclaration]
    _visibility: Visibility
    _documentation: Optional[StructuredDocumentation]

    def __init__(
        self,
        init: IrInitTuple,
        struct_definition: SolcStructDefinition,
        parent: SolidityAbc,
    ):
        super().__init__(init, struct_definition, parent)
        self._canonical_name = struct_definition.canonical_name
        # TODO scope
        self._visibility = struct_definition.visibility

        self._members = []
        for member in struct_definition.members:
            self._members.append(VariableDeclaration(init, member, self))
        self._documentation = (
            StructuredDocumentation(init, struct_definition.documentation, self)
            if struct_definition.documentation is not None
            else None
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for member in self._members:
            yield from member

    def _parse_name_location(self) -> Tuple[int, int]:
        IDENTIFIER = r"[a-zA-Z$_][a-zA-Z0-9$_]*"
        STRUCT_RE = re.compile(
            r"^\s*struct\s+(?P<name>{identifier})".format(identifier=IDENTIFIER).encode(
                "utf-8"
            )
        )

        byte_start = self._ast_node.src.byte_offset
        match = STRUCT_RE.match(self._source)
        assert match
        return byte_start + match.start("name"), byte_start + match.end("name")

    @property
    def parent(self) -> Union[ContractDefinition, SourceUnit]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def canonical_name(self) -> str:
        return self._canonical_name

    @property
    @lru_cache(maxsize=2048)
    def declaration_string(self) -> str:
        return (
            f"struct {self.name}"
            + " {\n"
            + ";\n".join(f"    {member.declaration_string}" for member in self._members)
            + ";\n}"
        )

    @property
    def members(self) -> Tuple[VariableDeclaration, ...]:
        """
        Returns:
            Tuple of member variable declarations.
        """
        return tuple(self._members)

    @property
    def documentation(self) -> Optional[StructuredDocumentation]:
        """
        Added in Solidity 0.8.20.
        Returns:
            [NatSpec](https://docs.soliditylang.org/en/latest/natspec-format.html) documentation string, if any.
        """
        return self._documentation



import json

from websocket import WebSocket, create_connection

from .abc import ProtocolAbc


class WebsocketProtocol(ProtocolAbc):
    _uri: str
    _timeout: float
    _ws: WebSocket

    def __init__(self, uri: str, timeout: float):
        self._uri = uri
        self._timeout = timeout

    def __enter__(self):
        self._ws = create_connection(
            self._uri, skip_utf8_validation=True, timeout=self._timeout
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._ws.close()

    def send_recv(self, data: str):
        self._ws.send(data)  # pyright: ignore reportGeneralTypeIssues
        return json.loads(self._ws.recv())



from dataclasses import dataclass
from typing import List


@dataclass
class UnknownEvent:
    topics: List[bytes]
    data: bytes


def read_from_memory(offset: int, length: int, memory: List) -> bytearray:
    start_block = offset // 32
    start_offset = offset % 32
    end_block = (offset + length) // 32
    end_offset = (offset + length) % 32

    if start_block == end_block:
        if start_block >= len(memory):
            return bytearray(length)
        return bytearray.fromhex(memory[start_block])[start_offset:end_offset]
    else:
        if start_block >= len(memory):
            ret = bytearray(32 - start_offset)
        else:
            ret = bytearray.fromhex(memory[start_block])[start_offset:]
        for i in range(start_block + 1, end_block):
            if i >= len(memory):
                ret += bytearray(32)
            else:
                ret += bytearray.fromhex(memory[i])
        if end_block >= len(memory):
            ret += bytearray(end_offset)
        else:
            ret += bytearray.fromhex(memory[end_block])[:end_offset]
        return ret



from __future__ import annotations

from typing import Iterator, Union

from woke.ir.ast import (
    SolcYulFunctionCall,
    SolcYulIdentifier,
    SolcYulIf,
    SolcYulLiteral,
)

from ..utils import IrInitTuple
from .abc import YulAbc, YulStatementAbc
from .block import YulBlock
from .function_call import YulFunctionCall
from .identifier import YulIdentifier
from .literal import YulLiteral


class YulIf(YulStatementAbc):
    """
    TBD
    """

    _parent: YulBlock
    _body: YulBlock
    _condition: Union[YulFunctionCall, YulIdentifier, YulLiteral]

    def __init__(self, init: IrInitTuple, if_statement: SolcYulIf, parent: YulAbc):
        super().__init__(init, if_statement, parent)
        self._body = YulBlock(init, if_statement.body, self)
        if isinstance(if_statement.condition, SolcYulFunctionCall):
            self._condition = YulFunctionCall(init, if_statement.condition, self)
        elif isinstance(if_statement.condition, SolcYulIdentifier):
            self._condition = YulIdentifier(init, if_statement.condition, self)
        elif isinstance(if_statement.condition, SolcYulLiteral):
            self._condition = YulLiteral(init, if_statement.condition, self)
        else:
            assert False, f"Unexpected type: {type(if_statement.condition)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._condition
        yield from self._body

    @property
    def parent(self) -> YulBlock:
        return self._parent

    @property
    def body(self) -> YulBlock:
        return self._body

    @property
    def condition(self) -> Union[YulFunctionCall, YulIdentifier, YulLiteral]:
        return self._condition



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, FrozenSet, Iterator, Optional, Set, Tuple, Union

if TYPE_CHECKING:
    from ..expressions.identifier import Identifier
    from ..meta.identifier_path import IdentifierPathPart
    from ..expressions.member_access import MemberAccess
    from ..statements.inline_assembly import ExternalReference
    from ..expressions.unary_operation import UnaryOperation
    from ..expressions.binary_operation import BinaryOperation

from woke.ir.abc import SolidityAbc
from woke.ir.ast import (
    SolcContractDefinition,
    SolcEnumDefinition,
    SolcEnumValue,
    SolcErrorDefinition,
    SolcEventDefinition,
    SolcFunctionDefinition,
    SolcModifierDefinition,
    SolcStructDefinition,
    SolcUserDefinedValueTypeDefinition,
    SolcVariableDeclaration,
)
from woke.ir.utils import IrInitTuple

SolcDeclarationUnion = Union[
    SolcContractDefinition,
    SolcEnumDefinition,
    SolcEnumValue,
    SolcErrorDefinition,
    SolcEventDefinition,
    SolcFunctionDefinition,
    SolcModifierDefinition,
    SolcStructDefinition,
    SolcUserDefinedValueTypeDefinition,
    SolcVariableDeclaration,
]


class DeclarationAbc(SolidityAbc, ABC):
    """
    Abstract base class for all Solidity declarations.
    """

    _name: str
    _name_location: Optional[Tuple[int, int]]
    _references: Set[
        Union[
            Identifier,
            IdentifierPathPart,
            MemberAccess,
            ExternalReference,
            UnaryOperation,
            BinaryOperation,
        ]
    ]

    def __init__(
        self, init: IrInitTuple, solc_node: SolcDeclarationUnion, parent: SolidityAbc
    ):
        super().__init__(init, solc_node, parent)
        self._name = solc_node.name
        if solc_node.name_location is None or solc_node.name_location.byte_offset < 0:
            self._name_location = None
        else:
            self._name_location = (
                solc_node.name_location.byte_offset,
                solc_node.name_location.byte_offset
                + solc_node.name_location.byte_length,
            )
        self._references = set()

    def register_reference(
        self,
        reference: Union[
            Identifier,
            IdentifierPathPart,
            MemberAccess,
            ExternalReference,
            UnaryOperation,
            BinaryOperation,
        ],
    ):
        self._references.add(reference)

    def unregister_reference(
        self,
        reference: Union[
            Identifier,
            IdentifierPathPart,
            MemberAccess,
            ExternalReference,
            UnaryOperation,
            BinaryOperation,
        ],
    ):
        self._references.remove(reference)

    def get_all_references(
        self, include_declarations: bool
    ) -> Iterator[
        Union[
            DeclarationAbc,
            Union[
                Identifier,
                IdentifierPathPart,
                MemberAccess,
                ExternalReference,
                UnaryOperation,
                BinaryOperation,
            ],
        ]
    ]:
        if include_declarations:
            yield self
        yield from self.references

    @abstractmethod
    def _parse_name_location(self) -> Tuple[int, int]:
        ...

    @property
    def name(self) -> str:
        """
        Returns:
            User-defined name of the declared object.
        """
        return self._name

    @property
    @abstractmethod
    def canonical_name(self) -> str:
        """
        !!! example
            `ContractName.StructName.FieldName` in the case of the `FieldName` [VariableDeclaration][woke.ir.declarations.variable_declaration.VariableDeclaration] declaration in the following example:
            ```solidity
            contract ContractName {
                struct StructName {
                    uint FieldName;
                }
            }
            ```
        Returns:
            Canonical name of the declared object.
        """
        ...

    @property
    @abstractmethod
    def declaration_string(self) -> str:
        """
        Declaration string that can be used for example in LSP hover. Does not include the declaration body, if any.
        Does not need to match the actual declaration string in the source code (may use a different order of keywords, for example).
        !!! example
            `:::solidity function foo(uint a, uint b) public payable virtual onlyOwner returns (uint)` of the [FunctionDefinition][woke.ir.declarations.function_definition.FunctionDefinition] declaration in the following example:
            ```solidity
            function foo(uint a, uint b) public onlyOwner virtual payable returns( uint ) {
                return a + b;
            }
            ```
        Returns:
            String representation of the declaration.
        """
        ...

    @property
    def name_location(self) -> Tuple[int, int]:
        """
        Similar to [byte_location][woke.ir.abc.IrAbc.byte_location], but returns the location of the declaration name in the source code.
        Returns:
            Tuple of the start and end byte offsets of the declaration name in the source code.
        """
        if self._name_location is None:
            self._name_location = self._parse_name_location()
        return self._name_location

    @property
    def references(
        self,
    ) -> FrozenSet[
        Union[
            Identifier,
            IdentifierPathPart,
            MemberAccess,
            ExternalReference,
            UnaryOperation,
            BinaryOperation,
        ]
    ]:
        """
        Returns:
            Set of all IR nodes referencing to this declaration.
        """
        return frozenset(self._references)



from __future__ import annotations

import logging
from collections import defaultdict
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Callable, DefaultDict, List, Optional, Set, Tuple
from urllib.error import HTTPError

from ipdb.__main__ import _init_pdb

from woke.config import WokeConfig
from woke.development.chain_interfaces import ChainInterfaceAbc
from woke.development.json_rpc import JsonRpcError
from woke.utils.file_utils import is_relative_to

if TYPE_CHECKING:
    from woke.testing.coverage import CoverageHandler


logger = logging.getLogger(__name__)


# must be declared before functions that use it because of a bug in Python (https://bugs.python.org/issue34939)
_exception_handler: Optional[Callable[[Exception], None]] = None
_exception_handled = False

_coverage_handler: Optional[CoverageHandler] = None

_config: Optional[WokeConfig] = None


def attach_debugger(e: Exception):
    global _exception_handled

    if _exception_handled:
        return
    _exception_handled = True

    import sys
    import traceback

    from woke.cli.console import console

    tb: Optional[TracebackType] = sys.exc_info()[2]
    assert tb is not None
    console.print_exception()

    frames = []

    for frame, lineno in traceback.walk_tb(tb):
        frames.append((frame, lineno))

    frames_up = 0
    for frame, lineno in reversed(frames):
        if is_relative_to(
            Path(frame.f_code.co_filename), Path.cwd()
        ) and not is_relative_to(
            Path(frame.f_code.co_filename), Path().cwd() / "pytypes"
        ):
            break
        frames_up += 1

    p = _init_pdb(
        commands=["up %d" % frames_up] if frames_up > 0 else [],
    )
    p.reset()
    p.interaction(None, tb)


def get_exception_handler() -> Optional[Callable[[Exception], None]]:
    return _exception_handler


def set_exception_handler(handler: Callable[[Exception], None]):
    global _exception_handler
    _exception_handler = handler


def reset_exception_handled():
    global _exception_handled
    _exception_handled = False


def get_config() -> WokeConfig:
    global _config
    if _config is None:
        _config = WokeConfig()
        _config.load_configs()
    return _config


def set_config(config: WokeConfig):
    global _config
    _config = config


def set_coverage_handler(coverage_handler: CoverageHandler):
    global _coverage_handler
    _coverage_handler = coverage_handler


def get_coverage_handler() -> Optional[CoverageHandler]:
    return _coverage_handler


class ChainInterfaceManager:
    _chain_interfaces: DefaultDict[Tuple, List[Tuple[ChainInterfaceAbc, Optional[str]]]]
    _free_chain_interfaces: DefaultDict[Tuple, Set[ChainInterfaceAbc]]

    def __init__(self):
        self._chain_interfaces = defaultdict(list)
        self._free_chain_interfaces = defaultdict(set)

    def get_or_create(
        self,
        uri: Optional[str],
        *,
        accounts: Optional[int],
        chain_id: Optional[int],
        fork: Optional[str],
        hardfork: Optional[str],
    ) -> ChainInterfaceAbc:
        params = (uri, accounts, chain_id, fork, hardfork)

        if len(self._free_chain_interfaces[params]) == 0:
            if uri is None:
                logger.debug(
                    "Launching chain with accounts=%s, chain_id=%s, fork=%s, hardfork=%s",
                    accounts,
                    chain_id,
                    fork,
                    hardfork,
                )
                chain_interface = ChainInterfaceAbc.launch(
                    get_config(),
                    accounts=accounts,
                    chain_id=chain_id,
                    fork=fork,
                    hardfork=hardfork,
                )
            else:
                if (
                    accounts is not None
                    or chain_id is not None
                    or fork is not None
                    or hardfork is not None
                ):
                    raise ValueError(
                        "Cannot specify accounts, chain_id, fork or hardfork when connecting to a running chain"
                    )
                logger.debug("Connecting to chain at %s", uri)
                chain_interface = ChainInterfaceAbc.connect(get_config(), uri)
        else:
            logger.debug(
                "Reusing chain with accounts=%s, chain_id=%s, fork=%s, hardfork=%s",
                accounts,
                chain_id,
                fork,
                hardfork,
            )
            chain_interface = self._free_chain_interfaces[params].pop()

        try:
            snapshot = chain_interface.snapshot()
        except (JsonRpcError, HTTPError):
            snapshot = None

        self._chain_interfaces[params].append((chain_interface, snapshot))
        return chain_interface

    def free(self, chain_interface: ChainInterfaceAbc) -> None:
        snapshot_reverted = False
        params = None

        for chain_params, chain_interfaces in self._chain_interfaces.items():
            for i, (c, snapshot) in enumerate(chain_interfaces):
                if c == chain_interface:
                    params = chain_params
                    self._chain_interfaces[chain_params].pop(i)
                    try:
                        if snapshot is not None and chain_interface.revert(snapshot):
                            snapshot_reverted = True
                    except JsonRpcError:
                        pass
                    break

        if snapshot_reverted and params is not None:
            logger.debug(
                "Freeing chain with uri=%s, accounts=%s, chain_id=%s, fork=%s, hardfork=%s",
                *params,
            )
            self._free_chain_interfaces[params].add(chain_interface)
        else:
            if params is None:
                logger.debug("Unable to revert snapshot, closing chain")
            else:
                logger.debug(
                    "Unable to revert snapshot, closing chain with uri=%s, accounts=%s, chain_id=%s, fork=%s, hardfork=%s",
                    *params,
                )
            chain_interface.close()

    def close(self, chain_interface: ChainInterfaceAbc) -> None:
        chain_interface.close()

        for chain_params, chain_interfaces in self._chain_interfaces.items():
            for i, (c, _) in enumerate(chain_interfaces):
                if c == chain_interface:
                    chain_interfaces.pop(i)
                    logger.debug(
                        "Closed chain with uri=%s, accounts=%s, chain_id=%s, fork=%s, hardfork=%s",
                        *chain_params,
                    )
                    return
        logger.debug("Closed chain")

    def close_all(self) -> None:
        for _, chain_interfaces in self._chain_interfaces.items():
            for chain_interface, _ in chain_interfaces:
                chain_interface.close()
        for _, chain_interfaces in self._free_chain_interfaces.items():
            for chain_interface in chain_interfaces:
                chain_interface.close()
        self._chain_interfaces.clear()
        self._free_chain_interfaces.clear()


chain_interfaces_manager = ChainInterfaceManager()



import logging
from functools import lru_cache, reduce
from operator import or_
from typing import Iterator, List, Optional, Set, Tuple, Union

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.enums import (
    FunctionCallKind,
    GlobalSymbolsEnum,
    ModifiesStateFlag,
    StateMutability,
)

from ..ast import SolcFunctionCall
from ..declarations.error_definition import ErrorDefinition
from ..declarations.event_definition import EventDefinition
from ..declarations.function_definition import FunctionDefinition
from ..declarations.modifier_definition import ModifierDefinition
from ..declarations.struct_definition import StructDefinition
from ..declarations.variable_declaration import VariableDeclaration
from ..expressions.abc import ExpressionAbc
from ..expressions.function_call_options import FunctionCallOptions
from ..expressions.identifier import Identifier
from ..expressions.member_access import MemberAccess
from ..expressions.new_expression import NewExpression
from ..expressions.tuple_expression import TupleExpression
from ..utils import IrInitTuple

logger = logging.getLogger(__name__)


class FunctionCall(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcFunctionCall
    _parent: SolidityAbc  # TODO: make this more specific

    _arguments: List[ExpressionAbc]
    _expression: ExpressionAbc
    _kind: FunctionCallKind
    _names: List[str]
    _try_call: bool

    _recursion_lock: bool

    def __init__(
        self, init: IrInitTuple, function_call: SolcFunctionCall, parent: SolidityAbc
    ):
        super().__init__(init, function_call, parent)
        self._recursion_lock = False
        self._kind = function_call.kind
        self._names = list(function_call.names)
        self._try_call = function_call.try_call

        self._expression = ExpressionAbc.from_ast(init, function_call.expression, self)
        self._arguments = [
            ExpressionAbc.from_ast(init, argument, self)
            for argument in function_call.arguments
        ]

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for argument in self._arguments:
            yield from argument
        yield from self._expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def kind(self) -> FunctionCallKind:
        return self._kind

    @property
    def names(self) -> Tuple[str, ...]:
        return tuple(self._names)

    @property
    def try_call(self) -> bool:
        return self._try_call

    @property
    def expression(self) -> ExpressionAbc:
        return self._expression

    @property
    def arguments(self) -> Tuple[ExpressionAbc, ...]:
        return tuple(self._arguments)

    @property
    @lru_cache(maxsize=2048)
    def function_called(
        self,
    ) -> Optional[
        Union[
            EventDefinition,
            ErrorDefinition,
            FunctionDefinition,
            GlobalSymbolsEnum,
            StructDefinition,
            VariableDeclaration,
        ]
    ]:
        if self.kind == FunctionCallKind.TYPE_CONVERSION:
            return None

        node = self.expression
        while True:
            if isinstance(node, Identifier):
                referenced_declaration = node.referenced_declaration
                if isinstance(
                    referenced_declaration,
                    (
                        EventDefinition,
                        ErrorDefinition,
                        FunctionDefinition,
                        GlobalSymbolsEnum,
                        StructDefinition,
                        VariableDeclaration,
                    ),
                ):
                    return referenced_declaration
                else:
                    assert (
                        False
                    ), f"Unexpected function call referenced declaration type: {referenced_declaration}"
            elif isinstance(node, MemberAccess):
                referenced_declaration = node.referenced_declaration
                if isinstance(
                    referenced_declaration,
                    (
                        EventDefinition,
                        ErrorDefinition,
                        FunctionDefinition,
                        GlobalSymbolsEnum,
                        StructDefinition,
                        VariableDeclaration,
                    ),
                ):
                    return referenced_declaration
                else:
                    assert (
                        False
                    ), f"Unexpected function call referenced declaration type: {referenced_declaration}"
            elif isinstance(node, FunctionCall):
                node = node.expression
                while isinstance(
                    node, MemberAccess
                ) and node.referenced_declaration in {
                    GlobalSymbolsEnum.FUNCTION_VALUE,
                    GlobalSymbolsEnum.FUNCTION_GAS,
                }:
                    node = node.expression
            elif isinstance(node, FunctionCallOptions):
                node = node.expression
            elif isinstance(node, NewExpression):
                return None
            elif isinstance(node, TupleExpression):
                if len(node.components) != 1:
                    assert (
                        False
                    ), f"Unexpected function call child node: {node}\n{self.source}"
                node = node.components[0]
            else:
                assert (
                    False
                ), f"Unexpected function call child node: {node}\n{self.source}"

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        if self.kind == FunctionCallKind.TYPE_CONVERSION:
            return self.expression.is_ref_to_state_variable
        return False

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        if self._recursion_lock:
            return set()
        self._recursion_lock = True
        ret = self.expression.modifies_state | reduce(
            or_, (arg.modifies_state for arg in self.arguments), set()
        )

        if self.kind == FunctionCallKind.FUNCTION_CALL:
            called_function = self.function_called
            if called_function in {
                GlobalSymbolsEnum.SELFDESTRUCT,
                GlobalSymbolsEnum.SUICIDE,
            }:
                ret |= {(self, ModifiesStateFlag.SELFDESTRUCTS)}
            elif called_function in {
                GlobalSymbolsEnum.ADDRESS_TRANSFER,
                GlobalSymbolsEnum.ADDRESS_SEND,
            }:
                ret |= {(self, ModifiesStateFlag.SENDS_ETHER)}
            elif called_function == GlobalSymbolsEnum.ADDRESS_CALL:
                ret |= {(self, ModifiesStateFlag.PERFORMS_CALL)}
            elif called_function == GlobalSymbolsEnum.ADDRESS_DELEGATECALL:
                ret |= {(self, ModifiesStateFlag.PERFORMS_DELEGATECALL)}
            elif (
                called_function
                in {GlobalSymbolsEnum.ARRAY_PUSH, GlobalSymbolsEnum.ARRAY_POP}
                and self.expression.is_ref_to_state_variable
            ):
                ret |= {(self, ModifiesStateFlag.MODIFIES_STATE_VAR)}
            elif called_function == GlobalSymbolsEnum.FUNCTION_VALUE:
                ret |= {(self, ModifiesStateFlag.SENDS_ETHER)}
            elif isinstance(called_function, FunctionDefinition):
                if called_function.state_mutability in {
                    StateMutability.PURE,
                    StateMutability.VIEW,
                }:
                    pass
                elif called_function.body is not None:
                    ret |= called_function.body.modifies_state
                    for modifier in called_function.modifiers:
                        modifier_def = modifier.modifier_name.referenced_declaration
                        assert isinstance(modifier_def, ModifierDefinition)
                        if modifier_def.body is not None:
                            ret |= modifier_def.body.modifies_state
                elif called_function.state_mutability == StateMutability.NONPAYABLE:
                    ret |= {
                        (
                            self,
                            ModifiesStateFlag.CALLS_UNIMPLEMENTED_NONPAYABLE_FUNCTION,
                        )
                    }
                elif called_function.state_mutability == StateMutability.PAYABLE:
                    ret |= {
                        (self, ModifiesStateFlag.CALLS_UNIMPLEMENTED_PAYABLE_FUNCTION)
                    }
                else:
                    assert False
        self._recursion_lock = False
        return ret



import copy
import logging
from collections import deque
from typing import Deque, Dict, List, Optional, Set, Tuple, Union

import networkx as nx

import woke.ir.types as types
from woke.ir import (
    Assignment,
    ContractDefinition,
    ElementaryTypeNameExpression,
    ExpressionAbc,
    ExpressionStatement,
    FunctionCall,
    FunctionCallOptions,
    FunctionDefinition,
    Identifier,
    InlineAssembly,
    MemberAccess,
    Return,
    VariableDeclaration,
    YulAbc,
    YulAssignment,
    YulIdentifier,
)
from woke.ir.enums import FunctionCallKind, GlobalSymbolsEnum

from .cfg import CfgBlock

logger = logging.getLogger(__name__)


def pair_function_call_arguments(
    definition: FunctionDefinition, call: FunctionCall
) -> Tuple[Tuple[VariableDeclaration, ExpressionAbc], ...]:
    if len(definition.parameters.parameters) == len(call.arguments):
        return tuple(zip(definition.parameters.parameters, call.arguments))
    elif len(definition.parameters.parameters) == len(call.arguments) + 1:
        # using for
        node = call.expression
        if isinstance(node, FunctionCallOptions):
            node = node.expression
        if isinstance(node, MemberAccess):
            node = node.expression
        return ((definition.parameters.parameters[0], node),) + tuple(
            zip(definition.parameters.parameters[1:], call.arguments)
        )
    else:
        raise ValueError(
            f"{definition.name} has {len(definition.parameters.parameters)} parameters but called with {len(call.arguments)} arguments"
        )


def get_all_base_and_child_functions(
    func: Union[FunctionDefinition, VariableDeclaration]
) -> Set[Union[FunctionDefinition, VariableDeclaration]]:
    ret = {func}
    queue: Deque[Union[FunctionDefinition, VariableDeclaration]] = deque([func])

    while len(queue) > 0:
        func = queue.popleft()

        if isinstance(func, VariableDeclaration):
            for base in func.base_functions:
                if base not in ret:
                    ret.add(base)
                    queue.append(base)
        elif isinstance(func, FunctionDefinition):
            for base in func.base_functions:
                if base not in ret:
                    ret.add(base)
                    queue.append(base)
            for child in func.child_functions:
                if child not in ret:
                    ret.add(child)
                    queue.append(child)
    return ret


def expression_is_global_symbol(
    expr: ExpressionAbc, global_symbol: GlobalSymbolsEnum
) -> bool:
    if isinstance(expr, MemberAccess):
        if expr.referenced_declaration == global_symbol:
            logger.debug(f"Expression is {global_symbol}: {expr.source}")
            return True
    elif (
        isinstance(expr, FunctionCall)
        and expr.kind == FunctionCallKind.TYPE_CONVERSION
        and len(expr.arguments) == 1
    ):
        # payable(msg.sender) for example
        conversion = expr.expression
        if isinstance(conversion, ElementaryTypeNameExpression):
            t = conversion.type
            if isinstance(t, types.Type) and isinstance(t.actual_type, types.Address):
                return expression_is_global_symbol(expr.arguments[0], global_symbol)
    elif isinstance(expr, FunctionCall) and isinstance(
        expr.function_called, FunctionDefinition
    ):
        func = expr.function_called
        assert isinstance(func, FunctionDefinition)
        if len(func.return_parameters.parameters) == 1:
            returns = []
            if func.body is None:
                return False

            for statement in func.body.statements_iter():
                if isinstance(statement, Return):
                    returns.append(statement)
            if all(
                expression_is_global_symbol(return_.expression, global_symbol)
                for return_ in returns
                if return_.expression is not None
            ):
                logger.debug(f"Expression is {global_symbol}: {expr.source}")
                return True

    logger.debug(f"Expression is NOT {global_symbol}: {expr.source}")
    return False


def get_variable_declarations_from_expression(
    expr: ExpressionAbc,
) -> Tuple[Optional[VariableDeclaration], ...]:
    if isinstance(expr, Identifier):
        ref = expr.referenced_declaration
        if isinstance(ref, VariableDeclaration):
            return (ref,)
    elif isinstance(expr, MemberAccess):
        ref = expr.referenced_declaration
        if isinstance(ref, VariableDeclaration):
            return get_variable_declarations_from_expression(expr.expression) + (ref,)
    elif isinstance(expr, FunctionCall) and isinstance(
        expr.function_called, (FunctionDefinition, VariableDeclaration)
    ):
        func = expr.function_called

        func_expr = expr.expression
        if isinstance(func_expr, Identifier):
            ret: Tuple[Optional[VariableDeclaration], ...] = tuple()
        elif isinstance(func_expr, MemberAccess):
            ret = get_variable_declarations_from_expression(func_expr.expression)
        else:
            return (None,)

        if isinstance(func, VariableDeclaration):
            return ret + (func,)
        elif isinstance(func, FunctionDefinition):
            if func.body is None:
                implementations = get_function_implementations(func)
                if len(implementations) == 1:
                    if isinstance(implementations[0], VariableDeclaration):
                        return ret + (implementations[0],)
                    elif isinstance(implementations[0], FunctionDefinition):
                        func = implementations[0]
                        assert func.body is not None
                    else:
                        return (None,)
                else:
                    return (None,)

            returns = []
            for statement in func.body.statements_iter():
                if isinstance(statement, Return):
                    returns.append(statement)
            if len(returns) == 1 and returns[0].expression is not None:
                return ret + get_variable_declarations_from_expression(
                    returns[0].expression
                )

    return (None,)


def get_function_implementations(
    function: FunctionDefinition,
) -> Tuple[Union[FunctionDefinition, VariableDeclaration], ...]:
    ret = set()
    visited: Set[Union[FunctionDefinition, VariableDeclaration]] = {function}

    queue = deque([function])
    while len(queue) > 0:
        func = queue.popleft()
        if func.implemented:
            ret.add(func)

        for f in func.child_functions:
            if f not in visited:
                visited.add(f)
                if isinstance(f, VariableDeclaration):
                    ret.add(f)
                elif isinstance(f, FunctionDefinition):
                    queue.append(f)
    return tuple(ret)


def get_function_definition_from_expression(
    expr: ExpressionAbc,
) -> Optional[FunctionDefinition]:
    fn_dec = expr
    while (
        fn_dec.parent
        and not isinstance(fn_dec, FunctionDefinition)
        and not isinstance(fn_dec, ContractDefinition)
    ):
        fn_dec = fn_dec.parent
    if (
        fn_dec is not None
        and isinstance(fn_dec, FunctionDefinition)
        and fn_dec.body is not None
    ):
        return fn_dec
    return None


def check_all_return_params_set(
    params: Tuple[VariableDeclaration],
    graph: nx.DiGraph,
    start_block: CfgBlock,
    end_block: CfgBlock,
) -> Tuple[
    bool, List[Dict[Tuple[Optional[VariableDeclaration]], Union[ExpressionAbc, YulAbc]]]
]:
    """
    Checks if all return parameters are set
    """
    out = []
    visited = set()
    all_set = _check_all_return_params_set(
        params, {}, set(), graph, start_block, end_block, out, visited
    )
    return all_set, out


def _check_all_return_params_set(
    params: Tuple[VariableDeclaration],
    solved_params: Dict[
        Tuple[Optional[VariableDeclaration]], Union[ExpressionAbc, YulAbc]
    ],
    solved_declarations: Set[VariableDeclaration],
    graph: nx.DiGraph,
    start_block: CfgBlock,
    end_block: CfgBlock,
    out: List[Dict[Tuple[Optional[VariableDeclaration]], Union[ExpressionAbc, YulAbc]]],
    visited: Set[CfgBlock],
) -> bool:
    if start_block in visited:
        return False
    visited.add(start_block)
    if len(solved_declarations) == len(params):
        out.append(solved_params)
        return True
    if start_block == end_block:
        out.append(solved_params)
        return False

    for stmt in reversed(start_block.statements):
        if isinstance(stmt, ExpressionStatement) and isinstance(
            stmt.expression, Assignment
        ):
            assigned_vars = []
            assigned_params = set()
            for vars in stmt.expression.assigned_variables:
                if vars is None:
                    continue
                for decls in vars:
                    for vd in decls:
                        if (
                            isinstance(vd, VariableDeclaration)
                            and vd in params
                            and vd not in solved_declarations
                        ):
                            assigned_vars.append(vd)
                            assigned_params.add(vd)
                        else:
                            assigned_vars.append(None)
            if len(assigned_params) > 0:
                solved_declarations.update(assigned_params)
                solved_params[tuple(assigned_vars)] = stmt.expression.right_expression
        elif isinstance(stmt, InlineAssembly) and stmt.yul_block is not None:
            for yul_stmt in stmt.yul_block.statements:
                if isinstance(yul_stmt, YulAssignment):
                    assigned_vars = []
                    assigned_params = set()
                    for var in yul_stmt.variable_names:
                        if (
                            isinstance(var, YulIdentifier)
                            and var.external_reference is not None
                            and isinstance(
                                var.external_reference.referenced_declaration,
                                VariableDeclaration,
                            )
                            and var.external_reference.referenced_declaration in params
                        ):
                            assigned_vars.append(
                                var.external_reference.referenced_declaration
                            )
                            assigned_params.add(
                                var.external_reference.referenced_declaration
                            )
                        else:
                            assigned_vars.append(None)
                    if len(assigned_params) > 0:
                        solved_declarations.update(assigned_params)
                        solved_params[tuple(assigned_vars)] = yul_stmt.value

    all_set = True
    for next_start in graph.predecessors(start_block):
        if next_start in visited:
            continue
        if not _check_all_return_params_set(
            params,
            copy.copy(solved_params),
            copy.copy(solved_declarations),
            graph,
            next_start,
            end_block,
            out,
            visited,
        ):
            all_set = False
    return all_set
